import importlib.metadata
import json
from logging.config import dictConfig
from typing import Any

from vllm.envs import VLLM_CONFIGURE_LOGGING, VLLM_LOGGING_CONFIG_PATH
from vllm.logger import DEFAULT_LOGGING_CONFIG

__version__ = importlib.metadata.version("vllm_spyre")


def register():
    """Register the Spyre platform."""
    return "vllm_spyre.platform.SpyrePlatform"


def _init_logging():
    """Setup logging, extending from the vLLM logging config"""
    config = dict[str, Any]()

    if VLLM_CONFIGURE_LOGGING:
        config = {**DEFAULT_LOGGING_CONFIG}

    if VLLM_LOGGING_CONFIG_PATH:
        # Error checks must be done already in vllm.logger.py
        with open(VLLM_LOGGING_CONFIG_PATH, encoding="utf-8") as file:
            config = json.loads(file.read())

    if VLLM_CONFIGURE_LOGGING:
        # Copy the vLLM logging configurations for our package
        if "vllm_spyre" not in config["formatters"]:
            if "vllm" in config["formatters"]:
                config["formatters"]["vllm_spyre"] = config["formatters"]["vllm"]
            else:
                config["formatters"]["vllm_spyre"] = DEFAULT_LOGGING_CONFIG["formatters"]["vllm"]

        if "vllm_spyre" not in config["handlers"]:
            if "vllm" in config["handlers"]:
                handler_config = config["handlers"]["vllm"]
            else:
                handler_config = DEFAULT_LOGGING_CONFIG["handlers"]["vllm"]
            handler_config["formatter"] = "vllm_spyre"
            config["handlers"]["vllm_spyre"] = handler_config

        if "vllm_spyre" not in config["loggers"]:
            if "vllm" in config["loggers"]:
                logger_config = config["loggers"]["vllm"]
            else:
                logger_config = DEFAULT_LOGGING_CONFIG["loggers"]["vllm"]
            logger_config["handlers"] = ["vllm_spyre"]
            config["loggers"]["vllm_spyre"] = logger_config

    dictConfig(config)


_init_logging()



"""Tests for our environment configs"""

import os

import pytest

from vllm_spyre import envs

pytestmark = pytest.mark.cpu


def test_env_vars_are_cached(monkeypatch):
    monkeypatch.setenv("VLLM_SPYRE_NUM_CPUS", "42")
    assert envs.VLLM_SPYRE_NUM_CPUS == 42

    # Future reads don't query the environment every time, so this should not
    # return the updated value
    monkeypatch.setenv("VLLM_SPYRE_NUM_CPUS", "77")
    assert envs.VLLM_SPYRE_NUM_CPUS == 42


def test_env_vars_override(monkeypatch):
    monkeypatch.setenv("VLLM_SPYRE_NUM_CPUS", "42")
    assert envs.VLLM_SPYRE_NUM_CPUS == 42

    # This override both sets the environment variable and updates our cache
    envs.override("VLLM_SPYRE_NUM_CPUS", "77")
    assert envs.VLLM_SPYRE_NUM_CPUS == 77
    assert os.getenv("VLLM_SPYRE_NUM_CPUS") == "77"


def test_env_vars_override_with_bad_value(monkeypatch):
    monkeypatch.setenv("VLLM_SPYRE_NUM_CPUS", "42")
    assert envs.VLLM_SPYRE_NUM_CPUS == 42

    # envs.override ensures the value can be parsed correctly
    with pytest.raises(ValueError, match=r"invalid literal for int"):
        envs.override("VLLM_SPYRE_NUM_CPUS", "notanumber")


def test_env_vars_override_for_invalid_config():
    with pytest.raises(ValueError, match=r"not a known setting"):
        envs.override("VLLM_SPYRE_NOT_A_CONFIG", "nothing")


def test_sendnn_decoder_backwards_compat(monkeypatch):
    # configuring the deprecated `sendnn_decoder` backend will swap to the new
    # `sendnn` backend instead
    monkeypatch.setenv("VLLM_SPYRE_DYNAMO_BACKEND", "sendnn_decoder")
    assert envs.VLLM_SPYRE_DYNAMO_BACKEND == "sendnn"



# SPDX-License-Identifier: Apache-2.0

import math
import os
from collections import deque
from typing import TYPE_CHECKING, Iterable, Union

from vllm.logger import init_logger
from vllm.v1.core.sched.scheduler import Scheduler
from vllm.v1.engine import EngineCoreOutputs
from vllm.v1.metrics.stats import PrefixCacheStats, SchedulerStats
from vllm.v1.outputs import ModelRunnerOutput
from vllm.v1.request import Request, RequestStatus

import vllm_spyre.envs as envs_spyre
from vllm_spyre.platform import SpyrePlatform
from vllm_spyre.v1.worker.spyre_model_runner import (
    CBSpyreModelRunnerOutput,
    CPSpyreModelRunnerOutput,
)

if TYPE_CHECKING:
    from vllm.v1.core.sched.output import SchedulerOutput
else:
    SchedulerOutput = None

logger = init_logger(__name__)

NO_WARMUP_FIT_STOP_REASON = "Request did not fit any warmup shape"


class SpyreScheduler(Scheduler):
    """Base class inheriting from the V1 scheduler to support static
    and continuous batching respecting AIU Spyre constraints."""

    def __init__(self, *args, **kwargs) -> None:
        # Initialize vLLM scheduler
        super().__init__(*args, **kwargs)
        self.model_config = self.vllm_config.model_config


class StaticBatchingSpyreScheduler(SpyreScheduler):
    """Support of static batching"""

    def __init__(self, *args, **kwargs) -> None:
        # Initialize SpyreScheduler
        super().__init__(*args, **kwargs)

        # Add our own state for handling Spyre constraints:
        # all warmup shapes that we can support
        self.spyre_warmup_shapes: tuple[dict[str, int], ...] = SpyrePlatform.get_warmup_shapes(
            self.scheduler_config
        )

    def schedule(self) -> SchedulerOutput:
        """This override adds constraints and then delegates most of the work
        to the base scheduler"""
        # First purge the full waiting queue into our holdback queue, preserving
        # priority, so that the base scheduler does not see them.
        # This lets us ensure that the set of requests scheduled have at least
        # one common warmup shape.
        holdback_queue: deque[Request] = deque()
        while self.waiting:
            holdback_queue.append(self.waiting.popleft())

        # store requests which don't fit the warmup shapes of the current batch
        skip_queue: deque[Request] = deque()

        # If no requests are currently running, we can now release requests back
        # into the waiting queue in priority order for the scheduler to prefill.
        # These must share a common warmup shape
        if len(self.running) == 0:
            # Make a copy of the warmup shapes
            available_warmup_shapes = list(self.spyre_warmup_shapes)
            last_available_warmup_shapes = available_warmup_shapes

            while holdback_queue:
                request = holdback_queue[0]

                # prune the possible shapes to only those that fit this request
                # and the growing batch size
                available_warmup_shapes = self._get_matching_warmup_shapes(
                    request=request,
                    warmup_shapes=available_warmup_shapes,
                    current_batch_size=len(self.waiting),
                )

                if len(available_warmup_shapes) > 0:
                    # There is still at least one valid shape, so add to the
                    # waiting queue
                    self.waiting.append(holdback_queue.popleft())
                    # remember the available warmup shapes of the current batch
                    last_available_warmup_shapes = available_warmup_shapes
                else:
                    # calculating the max possible batch size among the
                    # available warmup shapes of the scheduled requests
                    max_batch = max([d["batch_size"] for d in last_available_warmup_shapes])

                    # if there is potential space in the batch but the current
                    # request does not fit, skip it and try with the next
                    if len(self.waiting) < max_batch:
                        available_warmup_shapes = last_available_warmup_shapes
                        skip_queue.append(holdback_queue.popleft())
                    else:
                        # If the batch is full, we exit the loop here
                        break

            logger.debug(
                "Scheduling a new batch of %d requests, holding back %d requests",
                len(self.waiting),
                len(holdback_queue),
            )
        else:
            logger.debug("Scheduling a running batch of %d requests", len(self.running))

        # delegate to super of SpyreScheduler: base V1 Scheduler
        outputs = super(SpyreScheduler, self).schedule()

        # first move skipped and then unscheduled requests back
        # to the waiting queue, preserving priority
        while skip_queue:
            self.waiting.append(skip_queue.popleft())

        while holdback_queue:
            self.waiting.append(holdback_queue.popleft())

        return outputs

    def _get_matching_warmup_shapes(
        self, request: Request, warmup_shapes: list[dict[str, int]], current_batch_size: int
    ) -> list[dict[str, int]]:
        """Return the subset of shapes that match this request"""
        max_tokens = 0
        if request.sampling_params is not None and request.sampling_params.max_tokens is not None:
            max_tokens = request.sampling_params.max_tokens

        return [
            shape
            for shape in warmup_shapes
            if request.num_prompt_tokens <= shape["prompt_length"]
            and max_tokens <= shape["new_tokens"]
            and current_batch_size < shape["batch_size"]
        ]


class ContinuousBatchingSpyreScheduler(SpyreScheduler):
    """Support of continuous batching"""

    # inherited from V1 base scheduler but mypy needs to know the type
    running: list[Request]

    def __init__(self, *args, **kwargs) -> None:
        # Initialize SpyreScheduler
        super().__init__(*args, **kwargs)
        self.tkv = 0
        self.n_free_blocks = 0
        self.block_size = SpyrePlatform.get_block_size()
        self.max_batch_tkv_limit = os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT", default="-1")
        assert self.max_batch_tkv_limit != "-1", (
            "Expecting the env var VLLM_DT_MAX_BATCH_TKV_LIMIT to be set in platform.py"
        )
        # cache for self.check_batch_tkv_limit() outer key: tuple(request_ids),
        # inner key: (request_id, max_batch_tkv_limit), value: (lower, upper)
        self._cache_check_batch_tkv_limit: dict[tuple, dict[tuple, tuple]] = {}

    def update_from_output(
        self,
        scheduler_output: SchedulerOutput,
        model_runner_output: ModelRunnerOutput,
    ) -> dict[int, EngineCoreOutputs]:
        # Need an instance of CBSpyreModelRunnerOutput which holds the tkv value
        assert isinstance(model_runner_output, CBSpyreModelRunnerOutput), (
            "Expecting an instance of CBSpyreModelRunnerOutput when doing continuous batching."
        )
        self.tkv = model_runner_output.tkv
        self.n_free_blocks = model_runner_output.n_free_blocks
        return super(SpyreScheduler, self).update_from_output(scheduler_output, model_runner_output)

    def schedule(self) -> "SchedulerOutput":
        """This override adds constraints and then delegates most of the work
        to the base scheduler

        To avoid additional specialization, some requests are held back from the
        base scheduler but are restored after.
        """
        # First purge the full waiting queue into our holdback queue, preserving
        # priority, so that the base scheduler does not see them.
        holdback_queue: deque[Request] = deque()
        while self.waiting:
            holdback_queue.append(self.waiting.popleft())

        # Check if new requests can be scheduled.
        while holdback_queue:
            if self.can_schedule(holdback_queue[0]):
                # Add request to the waiting queue
                self.waiting.append(holdback_queue.popleft())
            else:
                # Otherwise, we simply stop here so that the scheduler
                # can work with the batch we have
                break

        # Schedule Prefill and Decode separately
        if len(self.waiting) > 0:
            # For prefill, hide current decodes from the scheduler
            running_holdback = self.running
            self.running = []
            logger.debug(
                "Scheduling a prefill step (%d prompt tokens), holding back %d requests",
                self.waiting[-1].num_prompt_tokens,  # ty: ignore[not-subscriptable]
                len(holdback_queue),
            )
        else:
            running_holdback = []
            logger.debug("Scheduling a decode step of %d requests", len(self.running))

        # delegate to super of SpyreScheduler: base V1 Scheduler
        outputs = super(SpyreScheduler, self).schedule()

        # restore holdbacks after running the base scheduler
        self.running = self.running + running_holdback
        while holdback_queue:
            self.waiting.append(holdback_queue.popleft())

        return outputs

    def can_schedule(self, request) -> bool:
        max_prompt_batch_size = 1
        max_context_len = self.model_config.max_model_len

        # running and waiting queues are both empty -> start a new batch
        # which can always be scheduled
        if len(self.running) + len(self.waiting) == 0:
            return True

        # check that there is space in the current decode batch
        cond1 = len(self.running) + len(self.waiting) < self.max_num_running_reqs

        # check that there is space in the prefill batch
        cond2 = len(self.waiting) < max_prompt_batch_size

        tkv = self.tkv

        # shift tkv if the prompt length exceeds the current tkv
        if request.num_prompt_tokens > self.tkv:
            tkv += (
                math.ceil((request.num_prompt_tokens - self.tkv) / self.block_size)
                * self.block_size
            )

        # check that the number of requested tokens can be served
        cond3 = request.max_tokens <= (max_context_len - tkv)

        # check condition for all other sequences in the current decode batch
        # if the tkv has been shifted
        if tkv > self.tkv:
            for req in self.running:
                n_generated_output_tokens = req.num_computed_tokens - req.num_prompt_tokens
                max_tokens_remaining = req.max_tokens - n_generated_output_tokens
                cond3_current = max_tokens_remaining <= (max_context_len - tkv)
                cond3 = cond3 and cond3_current
                # early exiting loop if condition is violated
                if not cond3:
                    return False

        # check that there are enough free blocks/pages remaining
        # Note: we only have to do check in case of a running batches
        # (not start_new_batch), because the minimal number of blocks covers
        # the context length for a single sequence, so tkv < block size is ok
        num_blocks_required = math.ceil((tkv + request.max_tokens - 1) / self.block_size)
        # optimization: subtract the padding blocks from the reserved blocks
        num_fully_padded_blocks = math.floor((tkv - request.num_prompt_tokens) / self.block_size)
        num_blocks_required -= num_fully_padded_blocks
        cond4 = num_blocks_required <= self.n_free_blocks

        # check that batch size x tkv is smaller than the max supported number
        cond5 = lambda: self.check_batch_tkv_limit(
            request=request,
            tkv=tkv,
            running=self.running,
            max_batch_tkv_limit=self.max_batch_tkv_limit,
        )

        return cond1 and cond2 and cond3 and cond4 and cond5()

    def check_batch_tkv_limit(self, request, tkv, running, max_batch_tkv_limit) -> bool:
        """
        Check whether adding a new sequence to the decode batch would violate
        Spyre's maximum batch volume constraint.

        In Spyre, the product of `batch_size` and the current `tkv`
        (tokens-per-sequence) must not exceed the limit defined by
        `VLLM_DT_MAX_BATCH_TKV_LIMIT`. Before scheduling a new sequence,
        we must ensure that this constraint will hold for all decoding
        steps that result from combining the new sequence with the currently
        running decode batch.

        This implementation:
        1. Computes the maximum possible `tkv` for each sequence in the
        decode batch.
        2. Sorts these values in ascending order.
        3. Iterates through them, stopping once the `tkv` of the new sequence.
        is reached. Remaining sequences do not need to be checked explicitly,
        since they were validated when they were added (by inductive reasoning).

        Note: drawing explaining the algorithm in more detail uploaded here:
        https://github.com/vllm-project/vllm-spyre/pull/363#issuecomment-3173605517
        """
        # checking if cached result can be used
        outer_key = tuple(r.request_id for r in running)  # decode batch changes
        inner_key = (request.request_id, max_batch_tkv_limit)  # new request changes
        cache = self._cache_check_batch_tkv_limit
        if (outer_key in cache) and (inner_key in cache[outer_key]):
            (lower, upper) = cache[outer_key][inner_key]
            if tkv <= lower or tkv >= upper:
                logger.debug(
                    "Cache hit function check_batch_tkv_limit: returning %s", str(tkv <= lower)
                )
                return tkv <= lower

        # Compute the effective token length of the new request
        new_req_max_tkv = tkv + request.max_tokens - 1

        # Compute token lengths for all running requests (decode batch)
        decode_req_max_tkvs = [
            tkv + req.max_tokens - 1 - (req.num_computed_tokens - req.num_prompt_tokens)
            for req in running
        ]
        # Sort decode requests token lengths in ascending order
        decode_req_max_tkvs.sort()

        # Initialize values
        batch_size = len(running) + 1
        max_batch_tkv = 0

        # Try adding the new request to the batch and check the max volume
        for decode_req_max_tkv in decode_req_max_tkvs:
            if new_req_max_tkv <= decode_req_max_tkv:
                # If the new request is shorter, it limits the batch volume
                max_batch_tkv = max(max_batch_tkv, batch_size * new_req_max_tkv)
                break
            else:
                # Otherwise, use the current (longer) request's volume
                max_batch_tkv = max(max_batch_tkv, batch_size * decode_req_max_tkv)
                # decrease batch_size by 1 as the current request finished
                batch_size -= 1

        return_value = max_batch_tkv <= int(max_batch_tkv_limit)

        if outer_key in cache:
            # decode batch has not changed
            if inner_key not in cache[outer_key]:
                # adding new request to present decode batch
                cache[outer_key][inner_key] = (-math.inf, math.inf)
        else:
            # decode batch has changed, empty the cache in place
            cache.clear()
            cache[outer_key] = {inner_key: (-math.inf, math.inf)}
            logger.debug(
                "Cleared cache of function check_batch_tkv_limit as the decode batch has changed."
            )

        # update lower bound (of acceptance) and upper bound (of rejection)
        (lower, upper) = cache[outer_key][inner_key]
        if return_value:
            lower = max(lower, tkv)
        else:
            upper = min(upper, tkv)
        assert lower < upper
        cache[outer_key][inner_key] = (lower, upper)
        logger.debug(
            "Saved cache of function check_batch_tkv_limit: %s",
            self._cache_check_batch_tkv_limit[outer_key][inner_key],
        )

        return return_value


class ChunkedPrefillSpyreScheduler(ContinuousBatchingSpyreScheduler):
    """
    Chunked-Prefill Scheduling policy

    The prefill vs. decode priority policy is the following:
        - Current prefill request priority: A new request cannot start prefill
           while another request's prefill is on-going

        - Prefill step interleaving: The prefill steps are interleaved with
            one decode step: as long as there are decoding requests, two
            prefill steps cannot be consecutive

        - General prefill priority: conditioned on interleaving constraint,
            prefill has priority over decode

        - No empty step: if a prefill step is prevented because it doesn't
            satisfy Spyre's specific constraints, a decode step is scheduled

    Spyre scheduling constraints:
        - Blocks constraint: all the blocks necessary to serve a request are
            allocated at the time of scheduling the first chunk of a chunked
            prefill. There must be enough blocks to accommodate the prompt and
            the maximum number of output tokens

        Note: all the remaining constraints need to be satisfied at the time
        of scheduling the last chunk of a chunked prefill

        - Prefill batch size: prefill batch is of size 1, only one request's
            chunked prefill can be scheduled at a time

        - Decode batch size: cannot have more than max_num_seqs running
            requests, including prefill and decode

        - Max model length constraint: the number of requested tokens must fit
            between the maximum TKV of all the running requests and the end of
            the model's context

        - Volumetric constraint: the total "surface" defined by the running
            requests should never exceed `VLLM_DT_MAX_BATCH_TKV_LIMIT`. See
            `check_batch_tkv_limit()` method for details.

        the surface defined by the maximum TKV of
            all the running requests and the number of running requests must
            not exceed the limit defined by `VLLM_DT_MAX_BATCH_TKV_LIMIT`
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.chunk_size = self.scheduler_config.max_num_batched_tokens

        # We want to keep track of requests for which the prefill is ongoing.
        # Theoretically, only one request can be prefilled at a time, but we
        # keep a list to be able to batch prefills in the future.
        self.ongoing_prefills: list[Request] = []

        # Prefills interleaving: if the feature flag is set, prefill operations
        # are interleaved with a decode step. This allows to minimize currently
        # decoding requests
        self.do_interleaving: bool = envs_spyre.VLLM_SPYRE_CP_INTERLEAVE_STEPS
        self.previous_step_was_prefill: bool = False

        # KV cache metrics
        # Prefix cache stats aggregated between each `make_stats` call
        self.prefix_cache_stats = PrefixCacheStats()
        # Gauge for kv cache usage, updated from every model runner output
        self.kv_cache_usage_percent = 0.0

    def update_from_output(self, scheduler_output, model_runner_output):
        assert isinstance(model_runner_output, CPSpyreModelRunnerOutput), (
            "Expecting an instance of CPSpyreModelRunnerOutput when doing chunked prefill."
        )

        # Update the correct num_computed_tokens value given left-padding and
        # prefix cache hit info
        for req in self.ongoing_prefills:
            # The number of computed tokens only need to be adapted when it is
            # the first chunk of a multi-chunk prefill
            is_first_chunk = req.num_computed_tokens <= self.chunk_size
            is_last_chunk = req.num_computed_tokens == req.num_prompt_tokens
            if is_first_chunk and not is_last_chunk:
                left_padding = model_runner_output.left_padding.get(req.request_id, 0)
                prefix_cache_len = model_runner_output.prefix_cache_hit_len.get(req.request_id, 0)

                req.num_computed_tokens = self.adjust_computed_tokens(
                    computed_tokens=req.num_computed_tokens,
                    left_padding=left_padding,
                    prefix_cache_len=prefix_cache_len,
                )

        # Remove completed prefills
        self.ongoing_prefills = [
            req for req in self.ongoing_prefills if req.num_computed_tokens < req.num_prompt_tokens
        ]

        # Update KV Cache info
        self.kv_cache_usage_percent = model_runner_output.kv_cache_usage
        prefix_cache_stats = model_runner_output.prefix_cache_stats
        if prefix_cache_stats is not None:
            # We don't use PrefixCacheStats.record because:
            # 1. It's not supported in vllm v0.11.0
            # 2. It could eventually be the case that requests > 1
            self.prefix_cache_stats.requests += prefix_cache_stats.requests
            self.prefix_cache_stats.queries += prefix_cache_stats.queries
            self.prefix_cache_stats.hits += prefix_cache_stats.hits

        return super().update_from_output(scheduler_output, model_runner_output)

    def adjust_computed_tokens(
        self, computed_tokens: int, left_padding: int, prefix_cache_len: int
    ) -> int:
        """
        Returns an adjusted `num_computed_tokens` given left padding and prefix
        cache hit info.
        """
        # The prefix cache length is already adjusted for left padding.
        # If it's bigger than the number of computed tokens, then we hit more
        # prefix cache than we scheduled.
        if prefix_cache_len > computed_tokens:
            assert (prefix_cache_len + left_padding) % self.chunk_size == 0
            return prefix_cache_len
        # Otherwise just account for the left padding
        return computed_tokens - left_padding

    def schedule(self) -> "SchedulerOutput":
        """
        The chunked prefill scheduling policy is enforced in this method, then
        delegates the final scheduling decision to the base scheduler

        To avoid additional specialization, some requests are held back from the
        base scheduler but are restored after
        """
        # First purge the full waiting queue into our holdback queue, preserving
        # priority, so that the base scheduler does not see them.
        holdback_queue: deque[Request] = deque()
        while self.waiting:
            holdback_queue.append(self.waiting.popleft())

        # Check if new requests can be scheduled for prefill
        while holdback_queue:
            if self.can_schedule_prefill(holdback_queue[0]):
                new_request = holdback_queue.popleft()
                # Remove structured_output_request
                # NB: SpyrePlatform.validate_request() removes structured_output
                # before the request gets here in most cases
                # TODO: We don't currently support structured output and it
                # breaks some assumptions the code makes. The problems is that
                # a structured output request will stay in waiting for multiple
                # iterations with status WAITING_FOR_FSM. To handle this
                # properly we need to exclude such requests from entering
                # ongoing_prefills but still pass them in the waiting queue to
                # the base scheduler to track the FSM initialization.
                if new_request.structured_output_request is not None:
                    logger.warning(
                        "Removing structured output from request: %s", new_request.request_id
                    )
                    new_request.structured_output_request = None
                    new_request.status = RequestStatus.WAITING

                logger.debug(
                    "Scheduling a new request (%d prompt tokens), holding back %d requests",
                    new_request.num_prompt_tokens,
                    len(holdback_queue),
                )

                # Add request to the waiting queue
                self.waiting.append(new_request)
            else:
                # Otherwise, we simply stop here so that the scheduler
                # can work with the batch we have
                break

        assert len(self.ongoing_prefills) <= 1, (
            "Only one request can be prefilled at a time, but got %d" % len(self.ongoing_prefills)
        )
        assert len(self.waiting) == 0 or len(self.ongoing_prefills) == 0, (
            "Cannot schedule new requests while another request prefill is ongoing."
        )
        assert all(r in self.running for r in self.ongoing_prefills), (
            "Ongoing prefill requests must be in the running queue."
        )

        # Check ongoing prefills
        if self.ongoing_prefills:
            # Some running requests are currently being prefilled. We need to
            # separate them from currently decoding requests, and schedule
            # them separately. Either we schedule a chunked prefill step, or a
            # decoding step

            assert len(self.ongoing_prefills) == 1

            schedule_prefill = self.can_schedule_prefill(self.ongoing_prefills[0])

            if schedule_prefill:
                running_holdback = [r for r in self.running if r not in self.ongoing_prefills]
                self.running = self.ongoing_prefills
                self.previous_step_was_prefill = True
            else:
                self.running = [r for r in self.running if r not in self.ongoing_prefills]
                running_holdback = self.ongoing_prefills
                self.previous_step_was_prefill = False

        # Check new requests to prefill
        elif len(self.waiting) > 0:
            self.ongoing_prefills.extend(self.waiting)
            # Hide current decodes from the scheduler
            running_holdback = self.running
            self.running = []
            self.previous_step_was_prefill = True
        else:
            self.previous_step_was_prefill = False
            running_holdback = []

        # delegate to super of SpyreScheduler: base V1 Scheduler
        outputs = super(SpyreScheduler, self).schedule()

        # restore holdbacks after running the base scheduler
        self.running = self.running + running_holdback
        while holdback_queue:
            self.waiting.append(holdback_queue.popleft())

        # Log the scheduled tokens not at every step, but when doing chunked
        # prefill. These include decode steps during interleaving
        if self.ongoing_prefills or any(
            r.num_computed_tokens <= r.num_prompt_tokens + 1 for r in self.running
        ):
            logger.debug("Scheduled tokens in this step: %s", outputs.num_scheduled_tokens)
        return outputs

    def can_schedule_prefill(self, request: Request) -> bool:
        # running and waiting queues are both empty, we can start a new batch
        # which can always be scheduled
        if len(self.running) + len(self.waiting) == 0:
            return True

        if not self._has_scheduling_priority(request):
            return False

        return self._satisfies_constraints(request)

    def _satisfies_constraints(self, request: Request) -> bool:
        is_first_chunk = request.num_computed_tokens == 0
        is_last_chunk = (request.num_prompt_tokens - request.num_computed_tokens) <= self.chunk_size

        if not self.do_interleaving:
            # All the prefills are consecutive, so the first chunk has to
            # satisfy all the constraints, and we don't need to check them again
            # for subsequent chunks.
            if not is_first_chunk:
                return True

            return self._satisfies_first_chunk_constraints(
                request
            ) and self._satisfies_last_chunk_constraints(request)

        can_schedule = True
        if is_first_chunk:
            can_schedule = self._satisfies_first_chunk_constraints(request)

        if is_last_chunk:
            can_schedule = can_schedule and self._satisfies_last_chunk_constraints(request)

        return can_schedule

    def _satisfies_first_chunk_constraints(self, request: Request) -> bool:
        """First chunked prefill can be scheduled only if there is space in the
        input batch (cond1) and in the prefill batch (cond2)."""

        # TODO theoretically we could already do a chunked prefill even
        # if the decode batch is full, but the current implementation of input
        # batch doesn't allow to do so.
        num_running = len(self.running)
        cond1 = num_running + len(self.waiting) < self.max_num_running_reqs

        # check that there is space in the prefill batch
        max_prefill_batch_size = 1
        cond2 = len(self.waiting) < max_prefill_batch_size

        return cond1 and cond2

    def _satisfies_last_chunk_constraints(self, request: Request) -> bool:
        """Last chunked prefill can be scheduled only if there is enough space
        in the decode batch, and if all the other spyre-related conditions
        are satisfied."""
        decoding_requests = [r for r in self.running if r not in self.ongoing_prefills]
        max_context_len = self.model_config.max_model_len

        # check that there is space in the current decode batch
        num_running = len(decoding_requests)
        cond1 = num_running + len(self.waiting) < self.max_num_running_reqs

        # calculate new max tkv of the batch given the new sequence joins
        # considers all possible cases:
        # - prompt_len > self.tkv and fall into different blocks
        # - prompt_len and self.tkv fall within the same block
        # - prompt_len < self.tkv and fall into different blocks
        prompt_len = request.num_prompt_tokens
        n_blocks = math.floor(max(self.tkv, prompt_len) / self.block_size)
        new_req_tkv = n_blocks * self.block_size + prompt_len % self.block_size

        # check that the number of requested tokens can be served for the
        # new sequence (optimal condition)
        # note that the -1 comes from the token we generate during prefill
        cond2 = request.max_tokens - 1 <= (max_context_len - new_req_tkv)
        # check cond2 for all other sequences in the current decode batch
        for req in decoding_requests:
            # current tkv of the (left aligned) decode sequence
            dec_req_tkv = n_blocks * self.block_size + req.num_computed_tokens % self.block_size
            n_generated_output_tokens = req.num_computed_tokens - req.num_prompt_tokens
            max_tokens_remaining = req.max_tokens - n_generated_output_tokens
            # note that the -1 comes from the token we generate during prefill
            cond2_current = max_tokens_remaining - 1 <= (max_context_len - dec_req_tkv)
            cond2 = cond2 and cond2_current
            # early exiting loop if violated 2nd condition
            if not cond2:
                return False

        # check that batch size x tkv is smaller than the max supported number
        # Note: using max_tkv is a conservative upper bound here. For the
        # optimal check we need model runner to return per sequence tkvs
        cond3 = lambda: self.check_batch_tkv_limit_cp(
            request=request,
            new_req_tkv=new_req_tkv,
            n_blocks=n_blocks,
            running=decoding_requests,
            max_batch_tkv_limit=self.max_batch_tkv_limit,
        )

        return cond1 and cond2 and cond3()

    def _has_scheduling_priority(self, request):
        decoding_requests = [r for r in self.running if r not in self.ongoing_prefills]

        # If we do interleaving, then two consecutive prefill steps are
        # forbidden when there are decoding requests
        if self.do_interleaving and self.previous_step_was_prefill and len(decoding_requests) > 0:
            return False

        # Requests that are already prefilling are prioritized over new requests
        if request in self.ongoing_prefills:
            return True

        # We can start prefilling a new requests if we satisfy the maximum
        # number of concurrent prefills
        max_concurrent_prefills = 1
        num_prefills = len(self.waiting) + len(self.ongoing_prefills)
        return num_prefills < max_concurrent_prefills

    def check_batch_tkv_limit_cp(
        self, request, new_req_tkv, n_blocks, running, max_batch_tkv_limit
    ) -> bool:
        """
        Check whether adding a new sequence to the decode batch would violate
        Spyre's maximum batch volume constraint for chunked prefill.

        In Spyre, the product of `batch_size` and the current `tkv`
        (tokens-per-sequence) must not exceed the limit defined by
        `VLLM_DT_MAX_BATCH_TKV_LIMIT`. Before scheduling a new sequence,
        we must ensure that this constraint will hold for all decoding
        steps that result from combining the new sequence with the currently
        running decode batch.

        This implementation:
        1. Computes the maximum possible `tkv` for each sequence in the
        decode batch.
        2. Sorts these values in ascending order.
        3. Iterates through them, stopping once the `tkv` of the new sequence.
        is reached. Remaining sequences do not need to be checked explicitly,
        since they were validated when they were added (by inductive reasoning).

        Note: drawing explaining the algorithm in more detail uploaded here:
        https://github.com/vllm-project/vllm-spyre/pull/363#issuecomment-3173605517
        """

        # Compute the effective token length of the new request
        new_req_max_tkv = new_req_tkv + request.max_tokens - 1

        # Compute token lengths for all running requests (decode batch)
        decode_req_max_tkvs = []
        for req in running:
            # current tkv of the (left aligned) decode sequence
            dec_req_tkv = n_blocks * self.block_size + req.num_computed_tokens % self.block_size
            n_generated_output_tokens = req.num_computed_tokens - req.num_prompt_tokens
            dec_req_max_tkv = dec_req_tkv + (req.max_tokens - n_generated_output_tokens) - 1
            decode_req_max_tkvs.append(dec_req_max_tkv)

        # Sort decode requests token lengths in ascending order
        decode_req_max_tkvs.sort()

        # Initialize values
        # The request is already in the running queue if it has done a first
        # chunked prefill
        batch_size = len(running)
        if request not in running:
            batch_size += 1
        max_batch_tkv = 0

        # Try adding the new request to the batch and check the max volume
        for decode_req_max_tkv in decode_req_max_tkvs:
            if new_req_max_tkv <= decode_req_max_tkv:
                # If the new request is shorter, it limits the batch volume
                max_batch_tkv = max(max_batch_tkv, batch_size * new_req_max_tkv)
                break
            else:
                # Otherwise, use the current (longer) request's volume
                max_batch_tkv = max(max_batch_tkv, batch_size * decode_req_max_tkv)
                # decrease batch_size by 1 as the current request finished
                batch_size -= 1

        return max_batch_tkv <= int(max_batch_tkv_limit)

    def finish_requests(
        self,
        request_ids: Union[str, Iterable[str]],
        finished_status,
    ) -> None:
        """Handles removing finished requests from ongoing_prefills"""
        if isinstance(request_ids, str):
            request_ids = (request_ids,)

        # first defer to vLLM scheduler where validation is handled
        super(SpyreScheduler, self).finish_requests(
            request_ids=request_ids, finished_status=finished_status
        )

        self.ongoing_prefills = [
            r for r in self.ongoing_prefills if r.request_id not in request_ids
        ]

    def make_stats(self, *args, **kwargs) -> SchedulerStats | None:
        """Update the scheduler stats from the base scheduler.
        These are used in the vllm StatLoggers, which are responsible for
        reporting stats in logs and metrics.
        """
        base_stats = super().make_stats(*args, **kwargs)

        if base_stats is not None:
            base_stats.kv_cache_usage = self.kv_cache_usage_percent
            base_stats.prefix_cache_stats = self.prefix_cache_stats
            # Every time `make_stats` is called we reset the prefix cache stats.
            # This mimics how the base scheduler handles the kv cache stats
            self.prefix_cache_stats = PrefixCacheStats()

        return base_stats



import dataclasses
import json
import time
from datetime import datetime
from functools import wraps
from pathlib import Path

from vllm.config import VllmConfig
from vllm.logger import init_logger
from vllm.v1.engine import async_llm, llm_engine
from vllm.v1.metrics.loggers import StatLoggerBase, StatLoggerManager
from vllm.v1.metrics.stats import FinishedRequestStats, IterationStats, SchedulerStats

from vllm_spyre import envs as envs_spyre

try:
    from vllm.v1.metrics.stats import MultiModalCacheStats
except ImportError:
    # compatibility for vllm pre 0.11.1
    class MultiModalCacheStats:
        pass


logger = init_logger(__name__)


@dataclasses.dataclass
class PerfRecord:
    """A record for request_metrics.jsonl.
    Contains info about a single finished request"""

    # ISO timestamp w/ milliseconds
    timestamp: str
    # timing info
    engine_stats: FinishedRequestStats
    # time spent pre-emptied for other prefills
    prefill_interrupt_seconds: float
    # ITL calculated without the prefill interrupts
    decode_only_itl_seconds: float

    # key names to append with a time unit during json serialization
    _TIME_KEYS = [
        "e2e_latency",
        "queued_time",
        "prefill_time",
        "inference_time",
        "decode_time",
        "mean_time_per_output_token",
    ]

    def to_json(self) -> str:
        json_dict = dataclasses.asdict(self)

        # Flatten the engine stats into the top level
        engine_dict = json_dict.pop("engine_stats")
        json_dict.update(engine_dict)

        # add _seconds onto the timing info from the engine
        for k in self._TIME_KEYS:
            if k in json_dict:
                json_dict[k + "_seconds"] = json_dict.pop(k)

        return json.dumps(json_dict)


class FileStatLogger(StatLoggerBase):
    def __init__(self, vllm_config: VllmConfig, engine_index=0):
        self.enabled = envs_spyre.VLLM_SPYRE_PERF_METRIC_LOGGING_ENABLED

        perf_dir = Path(envs_spyre.VLLM_SPYRE_PERF_METRIC_LOGGING_DIR)
        if not perf_dir.exists():
            perf_dir.mkdir(parents=True)

        self.perf_file = (
            Path(envs_spyre.VLLM_SPYRE_PERF_METRIC_LOGGING_DIR) / "request_metrics.jsonl"
        )

        if self.enabled and engine_index == 0:
            logger.info(
                "Initializing vllm-spyre perf debug logger. Writing perf info to: %s",
                str(self.perf_file),
            )

        # Clear any old metrics out first
        if self.perf_file.exists():
            self.perf_file.unlink()
        self.perf_file.touch()

        self.iso_format = "%Y-%m-%dT%H:%M:%S.%f"

        self._prefill_tuples: list[tuple[float, float]] = []
        self._max_batch_size = vllm_config.scheduler_config.max_num_seqs
        self._last_ts: float = 0

        self.open_file_pointer = self.perf_file.open("a")

    def __del__(self):
        self.open_file_pointer.close()

    def record(
        self,
        scheduler_stats: SchedulerStats | None,
        iteration_stats: IterationStats | None,
        mm_cache_stats: MultiModalCacheStats | None = None,
        engine_idx: int = 0,
    ):
        if not self.enabled or engine_idx != 0:
            # Only log from rank 0
            return

        if iteration_stats is None:
            return

        if iteration_stats.num_prompt_tokens > 0:
            self._save_prefill_time(iteration_stats)
        self._last_ts = iteration_stats.iteration_timestamp

        if not iteration_stats.finished_requests:
            # Only log finished requests
            return

        # Convert float timestamp to human readable string
        text_timestamp = datetime.fromtimestamp(iteration_stats.iteration_timestamp).strftime(
            self.iso_format
        )[:-3]

        records_to_write: list[str] = []
        for r in iteration_stats.finished_requests:
            # Calculate some estimates to add to the engine stats
            estimated_prefill_interrupt = self.estimate_prefill_interrupt_lower_bound(r)

            estimated_decode_itl = (r.decode_time - estimated_prefill_interrupt) / max(
                r.num_generation_tokens - 1, 1
            )

            record = PerfRecord(
                timestamp=text_timestamp,
                engine_stats=r,
                decode_only_itl_seconds=estimated_decode_itl,
                prefill_interrupt_seconds=estimated_prefill_interrupt,
            )
            records_to_write.append(record.to_json())

        self.open_file_pointer.write("\n".join(records_to_write) + "\n")
        self.open_file_pointer.flush()

    def log_engine_initialized(self):
        pass

    def _save_prefill_time(self, iteration_stats: IterationStats):
        """If this iteration was a prefill, then save the a tuple of the current
        time and prefill time. This will be used later to estimate a lower bound
        of the amount of time that other sequences were
        interrupted for this prefill to happen.

        This is only relevant because the batching implementation has to pause
        the running batch of decoding sequences to prefill a single sequence.
        """
        maybe_prefill_time = iteration_stats.iteration_timestamp - self._last_ts
        # TTFT here includes queueing and we don't have access to the iteration
        # duration itself so we have to try to calculate our own prefill time.
        # If we calculate an interval that was less than the reported TTFT, then
        # use it as the prefill time
        maybe_prefill_time = min(maybe_prefill_time, iteration_stats.time_to_first_tokens_iter[0])

        # Tuple is (timestamp, prefill_time)
        self._prefill_tuples.append((iteration_stats.iteration_timestamp, maybe_prefill_time))
        if len(self._prefill_tuples) > 2 * self._max_batch_size:
            # Delete older prefills, we can't hold everything in memory
            # Not guaranteed to be lossless
            self._prefill_tuples.pop(0)

    def estimate_prefill_interrupt_lower_bound(
        self, finished_request: FinishedRequestStats
    ) -> float:
        """Returns a lower bound estimate on the time (in ms) that this request
        was interrupted for other requests to prefill to join the batch"""
        estimated_prefill_interrupt: float = 0

        # NB: use current time instead of iteration timestamp to ensure that we
        # exclude current request's prefill
        slop = 0.001
        decode_start_time = time.time() - finished_request.decode_time + slop
        for i in range(len(self._prefill_tuples)):
            if self._prefill_tuples[i][0] > decode_start_time:
                # Sum up all prefills past decode start time
                estimated_prefill_interrupt = sum(r[1] for r in self._prefill_tuples[i:])
                break
        return estimated_prefill_interrupt


def file_stat_logger_factory(config: VllmConfig, engine_index=0) -> FileStatLogger:
    """Factory method accepted by vllm engine initializers"""
    return FileStatLogger(config, engine_index)


def patch_async_llm_stat_loggers():
    """
    🌶️🌶️🌶️
    Platforms cannot alter the initialization of a vllm engine, and the
    `stat_loggers` parameter is not user-settable via `EngineArgs`.

    So we resort to patching the initialization of the StatsLoggerManager to
    inject our own stats logger. This _should_ also be compatible with versions
    of vllm prior to the addition of `stats_loggers` engine parameter.
    🌶️🌶️🌶️
    """
    logger.debug("Setting up perf logger injection")
    original_init = StatLoggerManager.__init__

    @wraps(original_init)
    def new_init(self, *args, **kwargs):
        logger.debug("Injecting vllm-spyre perf logger factory")
        if "custom_stat_loggers" not in kwargs or kwargs["custom_stat_loggers"] is None:
            kwargs["custom_stat_loggers"] = []

        kwargs["custom_stat_loggers"].append(file_stat_logger_factory)

        original_init(self, *args, **kwargs)

    async_llm.StatLoggerManager.__init__ = new_init  # ty: ignore[invalid-assignment]
    llm_engine.StatLoggerManager.__init__ = new_init  # ty: ignore[invalid-assignment]



"""Verification of handling prompt length exceeding warmup shapes

Run `python -m pytest tests/e2e/test_spyre_max_prompt_length.py`.
"""

import pytest
from llm_cache import get_cached_llm
from spyre_util import DecodeWarmupShapes, ModelInfo, create_text_prompt
from vllm import SamplingParams


@pytest.mark.parametrize(
    "warmup_shapes", [[(64, 20, 4)], [(64, 20, 4), (128, 20, 2)]]
)  # (prompt_length/new_tokens/batch_size)
def test_max_prompt_len_and_new_tokens(
    model: ModelInfo, warmup_shapes: DecodeWarmupShapes, backend: str, use_llm_cache, monkeypatch
) -> None:
    """
    Simple test that for static batching:
    - prompts cannot exceed the maximum prompt length of all warmup shapes
    - max_tokens cannot exceed the max new token length of the matching warmup
        shape

    These two cases are combined to reduce the cost of starting each `LLM`
    """
    # monkeypatch.setenv("VLLM_SPYRE_DYNAMO_BACKEND", backend)
    # patch_warmup_shapes(warmup_shapes, monkeypatch)

    max_prompt_length = max([t[0] for t in warmup_shapes])
    max_new_tokens = max([t[1] for t in warmup_shapes])
    # llm = LLM(model)

    llm = get_cached_llm(
        model=model,
        max_model_len=256,  # unused
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
        max_num_seqs=None,
        use_cb=False,
    )

    # Craft a request with a prompt that is slightly too long for the warmup
    # shape
    prompt = create_text_prompt(
        model,
        min_token_length=max_prompt_length,
        max_token_length=max_prompt_length + max_new_tokens - 1,
    )
    sampling_params = SamplingParams(max_tokens=1)

    with pytest.raises(ValueError, match="warmup"):
        results = llm.generate(prompts=[prompt], sampling_params=sampling_params)
        assert results[0].outputs[0].text == ""

    # Craft a request with a prompt that fits, but where too many tokens are
    # requested
    prompt = "hello"
    sampling_params = SamplingParams(max_tokens=max_new_tokens + 1)
    with pytest.raises(ValueError, match="warmup"):
        results = llm.generate(prompts=[prompt], sampling_params=sampling_params)
        assert results[0].outputs[0].text == ""



import openai
import pytest


def _check_result(client, model, max_tokens=8, temperature=0.0, n=1) -> None:
    completion = client.completions.create(
        model=model,
        prompt="Hello World!",
        max_tokens=max_tokens,
        temperature=temperature,
        n=n,
    )
    assert len(completion.choices) == n
    assert len(completion.choices[0].text) > 0


def test_openai_serving(
    remote_openai_server,
    model,
    warmup_shapes,
    backend,
    tp_size,
    mode,
    max_num_seqs,
    max_model_len,
):
    """Test online serving using the `vllm serve` CLI"""

    client = remote_openai_server.get_client()
    model = model.name

    _check_result(client, model, n=1)
    _check_result(client, model, temperature=1.0, n=2)

    # rest are SB tests
    if mode == "sb":
        return
    # Check some basic error handling as well. This is all done in one test
    # now to avoid server boot-up overhead to test each case.
    # To change this we'll need:
    # - A better way to share a server as a test fixture, or
    # - Much less overhead on server boot (e.g. cached compiled graphs)
    with pytest.raises(openai.APIError):
        # Prompt too long should raise
        long_prompt = "Hello " * 1000
        client.completions.create(model=model, prompt=long_prompt, max_tokens=500)

    # Short prompt under context length but requesting too many tokens for
    # the warmup shape should return an empty result
    try:
        client.completions.create(model=model, prompt="Hello World!", max_tokens=25)
    except openai.BadRequestError as e:
        assert "warmup" in str(e)



import pytest
from scheduling_utils import create_request_for_scheduler_test, random_prompt

from v1.worker.mock_model import InstrumentedModelRunner

from spyre_util import REFERENCE_MODELS


@pytest.mark.cpu
@pytest.mark.worker
@pytest.mark.prefix_caching
def test_block_sharing_for_2_chunks(
    monkeypatch: pytest.MonkeyPatch,
):
    model = REFERENCE_MODELS[InstrumentedModelRunner.DEFAULT_TEST_MODEL]
    pc_model_runner = InstrumentedModelRunner.build(
        monkeypatch,
        max_num_batched_tokens=128,
    )
    prompt = random_prompt(model=model, seed=0, length=192)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=False,
        generate_hf_results=False,
        block_hasher=pc_model_runner.request_block_hasher,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=False,
        generate_hf_results=False,
        block_hasher=pc_model_runner.request_block_hasher,
    )

    chunk_plan = pc_model_runner._plan_chunking(request1.request)
    pc_model_runner.verify_chunk_plan(
        chunk_plan=chunk_plan,
        chunk_count=2,
        padding_blocks=1,
    )

    kv_cache_manager = pc_model_runner.kv_cache_manager

    kv_cache_manager.allocate_new_blocks(request1.request.request_id, 192)
    kv_cache_manager.cache_blocks(request1.request, 192)
    kv_cache_manager.free(request1.request.request_id)

    chunk_plan = pc_model_runner._plan_chunking(request2.request)
    pc_model_runner.verify_chunk_plan(
        chunk_plan=chunk_plan,
        chunk_count=2,
        padding_blocks=1,
        usable_cache_blocks=1,
        total_cache_blocks=3,
    )


@pytest.mark.cpu
@pytest.mark.worker
@pytest.mark.prefix_caching
def test_multi_chunk_partial_match_misaligned(
    monkeypatch: pytest.MonkeyPatch,
):
    """Scenario where two sequences are scheduled which share a common
    prefix. The second sequence shares 254 tokens with the first sequence,
    which is less than two chunks. We can therefore reuse only one chunk
    (254 < 2*128 = 256). This leads to recomputation of the third block.

    p1 = [AB|CD|EF]
    p2 = [AB|CX|EF]

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 384,  max tokens = 2, step joining = 0
            * 1: len = 384, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    pc_model_runner = InstrumentedModelRunner.build(
        monkeypatch=monkeypatch,
        max_num_batched_tokens=128,
        # needed number of available blocks: 16 + padding block = 17
        # needs to be multiple of batch size (2): 17 -> 18
        available_blocks=18,
    )

    # twice the same seed for a sequence of length 384
    # the first sequence shares the same prefix of length 384 tokens
    # the second sequence shares the same prefix of length 254 tokens
    # hence sequence 1 shares the first 254 tokens with sequence 0

    model = REFERENCE_MODELS[InstrumentedModelRunner.DEFAULT_TEST_MODEL]
    prompt1 = random_prompt(model=model, seed=0, length=384)
    prompt2 = prompt1[0:254] + random_prompt(model=model, seed=0, length=384 - 254)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=False,
        generate_hf_results=False,
        block_hasher=pc_model_runner.request_block_hasher,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt2,
        use_golden_token_injection=False,
        generate_hf_results=False,
        block_hasher=pc_model_runner.request_block_hasher,
    )

    # Schedule chunk 0 of request 0
    model_runner_output_1 = pc_model_runner.execute_new_request(
        request=request1.request, tokens_to_schedule=128
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2]],
        slot_mappings=[[1, 2]],
    )
    pc_model_runner.verify_model_runner_output(
        model_runner_output_1,
        req_ids=["0"],
        num_sampled_token_ids=0,
        tkv=384,
        # actual number of available blocks: 18 - 1 (padding block) = 17
        n_free_blocks=11,  # 17 - 6 = 11
        left_padding={"0": 0},
        prefix_cache_hit_len={"0": 0},
    )

    # Schedule chunk 1 of request 0
    model_runner_output_2 = pc_model_runner.execute_running_requests(
        req_ids=["0"],
        num_computed_tokens=[128],
        tokens_to_schedule=[128],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2, 3, 4]],
        slot_mappings=[[3, 4]],
    )
    pc_model_runner.verify_model_runner_output(
        model_runner_output_2,
        req_ids=["0"],
        num_sampled_token_ids=0,
        tkv=384,
        n_free_blocks=11,
        left_padding={"0": 0},
        prefix_cache_hit_len={"0": 0},
    )

    # Schedule chunk 2 of request 0
    model_runner_output_3 = pc_model_runner.execute_running_requests(
        req_ids=["0"],
        num_computed_tokens=[256],
        tokens_to_schedule=[128],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2, 3, 4, 5, 6]],
        slot_mappings=[[5, 6]],
    )
    pc_model_runner.verify_model_runner_output(
        model_runner_output_3,
        req_ids=["0"],
        num_sampled_token_ids=1,
        tkv=384,
        n_free_blocks=11,
        left_padding={"0": 0},
    )

    # Schedule chunk 0 of request 1
    model_runner_output_4 = pc_model_runner.execute_new_request(
        request=request2.request, tokens_to_schedule=128
    )
    # chunk loaded from cache
    assert pc_model_runner.model.last_attn_metadata is None
    pc_model_runner.verify_model_runner_output(
        model_runner_output_4,
        req_ids=["1"],
        num_sampled_token_ids=0,
        tkv=384,
        n_free_blocks=8,  # 11 - (6 - 3 (prefixes)) = 8
        left_padding={"1": 0},
        prefix_cache_hit_len={"1": 128},
    )

    # Schedule chunk 1 of request 1
    model_runner_output_5 = pc_model_runner.execute_running_requests(
        req_ids=["1"],
        num_computed_tokens=[128],
        tokens_to_schedule=[128],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2, 3, 7]],
        slot_mappings=[[0, 7]],
    )  # <-- HERE: the block table and slot mapping
    # point to different places when a block is recomputed

    pc_model_runner.verify_model_runner_output(
        model_runner_output_5,
        req_ids=["1"],
        num_sampled_token_ids=0,
        tkv=384,
        n_free_blocks=8,
        left_padding={"1": 0},
        prefix_cache_hit_len={"1": 128},
    )

    # Schedule chunk 2 of request 1
    model_runner_output_6 = pc_model_runner.execute_running_requests(
        req_ids=["1"],
        num_computed_tokens=[256],
        tokens_to_schedule=[128],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2, 3, 7, 8, 9]],
        slot_mappings=[[8, 9]],
    )
    pc_model_runner.verify_model_runner_output(
        model_runner_output_6,
        req_ids=["1"],
        num_sampled_token_ids=1,
        tkv=384,
        n_free_blocks=8,
        left_padding={"1": 0},
    )

    # Schedule decodes of requests 0 and 1
    model_runner_output_7 = pc_model_runner.execute_running_requests(
        req_ids=["1", "0"],
        num_computed_tokens=[384, 384],
        tokens_to_schedule=[1, 1],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2, 3, 4, 5, 6, 10], [1, 2, 3, 7, 8, 9, 11]],
        slot_mappings=[[10], [11]],
        slot_slice=slice(0, 1),
    )
    pc_model_runner.verify_model_runner_output(
        model_runner_output_7,
        req_ids=["0", "1"],
        num_sampled_token_ids=2,
        tkv=385,
        n_free_blocks=6,  # 8 - 2 (both seqs need 1 block for decodes) = 6
        left_padding={"0": 0, "1": 0},
    )


@pytest.mark.cpu
@pytest.mark.worker
@pytest.mark.prefix_caching
def test_first_chunk_recomputation(
    monkeypatch: pytest.MonkeyPatch,
):
    """Scenario where two sequences are scheduled with 2 blocks
    each and a common 1 block prefix. Since chunk size is 4 times the block
    size, the first two blocks of the first chunk of each request will be
    padding blocks. In the second request, the third block of the chunk
    will be recomputed to prevent block table deduplication while the
    fourth block will be computed from scratch.

    p1 = [00AB]
    p2 = [00AC]

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 128,  max tokens = 2, step joining = 0
            * 1: len = 128, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    pc_model_runner = InstrumentedModelRunner.build(
        monkeypatch=monkeypatch,
        max_num_batched_tokens=256,
        # needed number of available blocks: 16 + padding block = 17
        # needs to be multiple of batch size (2): 17 -> 18
        available_blocks=18,
    )

    model = REFERENCE_MODELS[InstrumentedModelRunner.DEFAULT_TEST_MODEL]
    prompt1 = random_prompt(model=model, seed=0, length=128)
    prompt2 = prompt1[0:64] + random_prompt(model=model, seed=0, length=128 - 64)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=False,
        generate_hf_results=False,
        block_hasher=pc_model_runner.request_block_hasher,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt2,
        use_golden_token_injection=False,
        generate_hf_results=False,
        block_hasher=pc_model_runner.request_block_hasher,
    )

    # Schedule chunk 0 of request 0
    model_runner_output_1 = pc_model_runner.execute_new_request(
        request=request1.request, tokens_to_schedule=128
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[0, 0, 1, 2]],
        slot_mappings=[[0, 0, 1, 2]],
    )
    pc_model_runner.verify_model_runner_output(
        model_runner_output_1,
        req_ids=["0"],
        num_sampled_token_ids=1,
        tkv=128,
        # actual number of available blocks: 18 - 1 (padding block) = 17
        n_free_blocks=15,  # 17 - (4 - 2 (pads)) = 15
        left_padding={"0": 128},
    )

    # Schedule chunk 0 of request 1
    model_runner_output_2 = pc_model_runner.execute_new_request(
        request=request2.request, tokens_to_schedule=128
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[0, 0, 1, 3]],
        slot_mappings=[[0, 0, 0, 3]],
    )
    # ^This block table and slot mapping is the crux of this test.
    # The padding blocks align with the slot mapping pointing to block
    # 0. The third block is a cache hit, but has to be recomputed
    # because we're in the last chunk with a prefix hit. The fourth
    # block is not a prefix hit and has to be computed from scratch
    # in a new block.

    pc_model_runner.verify_model_runner_output(
        model_runner_output_2,
        req_ids=["1"],
        num_sampled_token_ids=1,
        tkv=128,
        n_free_blocks=14,  # 15 - (4 - 2 (pads) - 1 (prefix)) = 14
        left_padding={"1": 128},
    )

    # Schedule decodes of requests 0 and 1
    model_runner_output_3 = pc_model_runner.execute_running_requests(
        req_ids=["1", "0"],
        num_computed_tokens=[128, 128],
        tokens_to_schedule=[1, 1],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2, 4], [1, 3, 5]],
        slot_mappings=[[4], [5]],
        slot_slice=slice(0, 1),
    )
    pc_model_runner.verify_model_runner_output(
        model_runner_output_3,
        req_ids=["0", "1"],
        num_sampled_token_ids=2,
        tkv=129,
        n_free_blocks=12,  # 14 - 2 (both seqs need 1 block for decodes) = 12
        left_padding={"0": 0, "1": 0},
    )


@pytest.mark.cpu
@pytest.mark.worker
@pytest.mark.prefix_caching
def test_middle_chunk_recomputation_with_padding(
    monkeypatch: pytest.MonkeyPatch,
):
    """Scenario where two sequences are scheduled. The first one has
    8 blocks and the second 10, where the first 8 blocks are the same
    as in the first sequence. Since the chunk size is 4 times the block
    size, the first two blocks of the first chunk of the second request
    will be padding blocks. Since the second chunk of the second request
    is a full hit and is not the last chunk, it can be skipped. The third
    chunk has a cache hit of two blocks, which have to be recomputed to
    prevent duplicates in the block table while the two last blocks will
    be computed from scratch.

    This test also exercises the optimization where the tkv of existing
    requests is increased and padding blocks are added to the left in the
    decode batch to accommodate new requests with larger prompts.

    p1 = [ ABCD | EFGH ]
    p2 = [ 00AB | CDEF | GHIJ ]

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 512,  max tokens = 2, step joining = 0
            * 1: len = 640, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    pc_model_runner = InstrumentedModelRunner.build(
        monkeypatch=monkeypatch,
        max_num_batched_tokens=256,
        # needed number of available blocks: 32 + padding block = 33
        # needs to be multiple of batch size (2): 33 -> 34
        available_blocks=34,
        max_model_len=1024,
    )

    model = REFERENCE_MODELS[InstrumentedModelRunner.DEFAULT_TEST_MODEL]
    prompt1 = random_prompt(model=model, seed=0, length=512)
    prompt2 = prompt1[0:512] + random_prompt(model=model, seed=0, length=128)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=False,
        generate_hf_results=False,
        block_hasher=pc_model_runner.request_block_hasher,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt2,
        use_golden_token_injection=False,
        generate_hf_results=False,
        block_hasher=pc_model_runner.request_block_hasher,
    )

    # Schedule chunk 0 of request 0
    model_runner_output_1 = pc_model_runner.execute_new_request(
        request=request1.request, tokens_to_schedule=256
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2, 3, 4]],
        slot_mappings=[[1, 2, 3, 4]],
    )

    pc_model_runner.verify_model_runner_output(
        model_runner_output_1,
        req_ids=["0"],
        num_sampled_token_ids=0,
        tkv=512,
        # actual number of available blocks: 34 - 1 (padding block) = 33
        n_free_blocks=25,  # 33 - 8 = 25
        left_padding={"0": 0},
        prefix_cache_hit_len={"0": 0},
    )

    # Schedule chunk 1 of request 0
    model_runner_output_2 = pc_model_runner.execute_running_requests(
        req_ids=["0"],
        num_computed_tokens=[256],
        tokens_to_schedule=[256],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[1, 2, 3, 4, 5, 6, 7, 8]],
        slot_mappings=[[5, 6, 7, 8]],
    )
    pc_model_runner.verify_model_runner_output(
        model_runner_output_2,
        req_ids=["0"],
        num_sampled_token_ids=1,
        tkv=512,
        n_free_blocks=25,
        left_padding={"0": 0},
    )

    # Schedule chunk 0 of request 1
    model_runner_output_3 = pc_model_runner.execute_new_request(
        request=request2.request, tokens_to_schedule=256
    )
    # chunk loaded from cache, therefore the attn_metadata is not set
    assert pc_model_runner.model.last_attn_metadata is None

    pc_model_runner.verify_model_runner_output(
        model_runner_output_3,
        req_ids=["1"],
        num_sampled_token_ids=0,
        tkv=512,
        n_free_blocks=23,  # 25 - (12 - 2 (pads) - 8 (prefix)) = 23
        left_padding={"1": 128},
        prefix_cache_hit_len={"1": 384},
    )

    # Schedule chunk 2 of request 1 skipping the middle chunk execution because
    # it's loaded from cache. We can consider the prefix_cache_hit_len
    # above to have been computed
    model_runner_output_4 = pc_model_runner.execute_running_requests(
        req_ids=["1"],
        num_computed_tokens=[384],
        tokens_to_schedule=[256],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]],
        slot_mappings=[[0, 0, 9, 10]],
    )

    pc_model_runner.verify_model_runner_output(
        model_runner_output_4,
        req_ids=["1"],
        num_sampled_token_ids=1,
        tkv=640,
        n_free_blocks=23,
        left_padding={"1": 128},
    )

    # Schedule decodes of requests 0 and 1
    model_runner_output_5 = pc_model_runner.execute_running_requests(
        req_ids=["1", "0"],
        num_computed_tokens=[640, 512],
        tokens_to_schedule=[1, 1],
    )
    pc_model_runner.assert_block_tables_and_slot_mappings(
        block_tables=[[0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 11], [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12]],
        slot_mappings=[[11], [12]],
        slot_slice=slice(0, 1),
    )

    pc_model_runner.verify_model_runner_output(
        model_runner_output_5,
        req_ids=["0", "1"],
        num_sampled_token_ids=2,
        tkv=641,
        n_free_blocks=21,  # 23 - 2 (both seqs need 1 block for decodes) = 21
        left_padding={"0": 128, "1": 0},
    )



import torch
from fms.utils import serialization
from fms.utils.config import ModelConfig
from transformers import PretrainedConfig
from vllm.multimodal.inputs import (
    MultiModalBatchedField,
    MultiModalFeatureSpec,
    MultiModalFieldElem,
    MultiModalKwargsItem,
    PlaceholderRange,
)

from vllm_spyre.multimodal.mm_mappings import MMUtilsBase, MMWarmupInputs

# Extend the adapter as part of the head dim fix; this is needed to
# load 2b models correctly, but we do it here since this class is
# currently initialized only once and the adapter extension does not
# seem to be idempotent.
#
# NOTE: If this is made idempotent, we can move this into
# get_mm_specific_load_overrides(), since it's needed to load.
serialization.extend_adapter("llava_next", "hf", ["weight_expansion_for_mismatched_head_dim"])


class LlavaNextMMUtils(MMUtilsBase):
    @staticmethod
    def _validate_configs(fms_config: ModelConfig, hf_config: PretrainedConfig):
        """Ensure that configs are properly typed. Additional validation, e.g.,
        validating subconfig attrs should generally be done within subclasses.
        """
        MMUtilsBase._validate_configs(fms_config, hf_config)
        if hf_config.model_type != "llava_next" or hf_config.text_config.model_type != "granite":
            raise TypeError("llava next currently only supports granite LLMs!")

    def unwrap_mm_kv_cache_opts(self):
        """Unwrap options to be passed for the kv cache from the underlying
        text configs and return the resulting dictionary, which is used to
        .update() the common kv cache opts that don't need unwrapping.
        """
        kv_cache_specs = {}
        # NOTE: this is granite LLM specific, since the only llava next
        # variant supported in FMS is currently granite vision.
        kv_cache_specs["num_layers"] = self.hf_config.text_config.num_hidden_layers
        kv_cache_specs["head_dim"] = getattr(
            self.fms_config.text_config,
            "head_dim",
            self.hf_config.text_config.hidden_size
            // self.hf_config.text_config.num_attention_heads,
        )
        return kv_cache_specs

    @staticmethod
    def get_mm_specific_load_overrides(hf_config: PretrainedConfig):
        """Get any overrides needed for initializing the FMS model from the
        transformers config. For this model, we need to fix the head_dim, which
        currently surfaces as a problem for all 2b variants of granite 3.x LLMs
        when running through FMS.

        TODO: If additional variants of granite vision are added, or broader
        llava next support is added in FMS, handle it properly here.
        """
        return {
            "override_hf_pretrained_config": True,
            "text_config": {"head_dim": 128},
        }

    @staticmethod
    def get_maybe_mm_embeddings(
        fms_model: torch.nn.Module,
        input_ids: torch.Tensor,
        mm_features: list[MultiModalFeatureSpec],
        is_decode: bool,
    ) -> torch.Tensor:
        """Get the text or multimodal embeddings for Llava Next using
        the (potentially compiled) FMS model.
        """
        fms_kwargs = {"use_cache": True}
        mm_spec_keys = ["pixel_values", "image_sizes"]

        # Only merge multimodal features in prefill; nothing mm in decode
        if mm_features:
            assert not is_decode  # We never pass features in decode
            if len(mm_features) != 1:
                raise ValueError("Currently we assume we only embed one mm request at a time")
            mm_spec = mm_features[0].data
            if mm_spec is not None:
                # NOTE: This should be pretty safe as it's dependent on the
                # vLLM/HF processor objects, but we check it anyway to be safe
                # for now, since transformers 5.0 is just around the corner.
                if any(k not in mm_spec for k in mm_spec_keys):
                    raise KeyError(f"Llava Next requires kwargs: {mm_spec_keys}")

                fms_kwargs["pixel_values"] = mm_spec["pixel_values"].data
                image_sizes = mm_spec["image_sizes"].data

                # Careful about this; if it's 1D, we'll a tensor of shape
                # [x, y], which will break in a weird way in image packing,
                # since it assumes it's 2D and will get sad about getting
                # an int instead of an iterable
                if image_sizes.ndim == 1:
                    image_sizes = image_sizes.unsqueeze(0)
                fms_kwargs["image_sizes"] = image_sizes

        # The value of iteration does not matter for decode as long as it's > 0
        input_embeds, _ = fms_model.prepare_inputs_for_generation(
            iteration=0 if not is_decode else 1, input_ids=input_ids, kwargs=fms_kwargs
        )  # ty: ignore[call-non-callable]
        return input_embeds

    def get_warmup_inputs(self, req_count: int) -> MMWarmupInputs:
        """Get the inputs to the huggingface processor to create the warmup
        features or feature shapes.
        """
        # Warmup text is just an image token
        dummy_tokens = [self.hf_processor.decode(self.get_multimodal_token_id())]

        # number of image tokens only depends on shape;
        # using a smaller image here uses less context.
        tile_size = self.hf_config.vision_config.image_size
        side_dim = tile_size // 2
        dummy_img = torch.zeros((3, side_dim, side_dim), dtype=torch.uint8)

        proc_res = self.hf_processor(
            text=dummy_tokens,
            images=dummy_img,
            return_tensors="pt",
        )

        seq_len = proc_res.input_ids.shape[-1]
        # Get the input tokens and embeddings; currently embeddings are used,
        # but tokens are still required for the interfaces to be happy.
        warmup_input_ids = proc_res.input_ids.squeeze(0)
        emb_dim = self.hf_config.text_config.hidden_size
        warmup_embeds = torch.rand((seq_len, emb_dim))
        # Get the multimodal features spec
        warmup_mm_features = LlavaNextMMUtils._build_multimodal_spec(proc_res)

        return MMWarmupInputs(
            input_ids=[warmup_input_ids.tolist()] * req_count,
            input_embeds=[warmup_embeds] * req_count,
            mm_features=warmup_mm_features,
        )

    @staticmethod
    def _build_multimodal_spec(proc_res):
        """Given output of the processor on warmup data, build MM features"""

        # Squeeze down batch dim here; all token inputs are image tokens
        num_img_toks = proc_res.input_ids.shape[-1]

        # Multimodal features / feature spec
        mm_position = PlaceholderRange(offset=0, length=num_img_toks)
        mm_data = {
            "pixel_values": proc_res.pixel_values.squeeze(axis=0),
            "image_sizes": proc_res.image_sizes.squeeze(axis=0),
        }
        mm_fields = MultiModalKwargsItem(
            {
                mm_key: MultiModalFieldElem(
                    modality="image", key=mm_key, data=mm_data, field=MultiModalBatchedField()
                )
                for mm_key, mm_data in mm_data.items()
            }
        )

        return [
            MultiModalFeatureSpec(
                data=mm_fields,
                modality="image",
                identifier="MM-warmup-llava-next",
                mm_position=mm_position,
            )
        ]

    def get_multimodal_token_id(self) -> int:
        return self.hf_config.image_token_index



"""Verification of the correctness of the step-by-step execution of chunked
prefill. It does so by comparing, at every engine step (i.e. prefill or decode
iteration), a bunch of attributes. This allows a finer testing of the padding
and scheduling implementation.

Run `python -m pytest tests/e2e/test_spyre_cp_inference_steps.py`.
"""

import pytest
from scheduling_utils import check_scheduler_inference_steps
from spyre_util import ModelInfo


@pytest.mark.chunked_prefill
@pytest.mark.full_model
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [128])  # restricted to violate scheduler condition
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_prefill_tkv_too_big(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    set_random_seed,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where the requested prompt is too long for current tkv value

    Note that as we could prefill the prompt straight away, however,
    in this test the max model length is decreased to a value where
    the tkv of the decode batch would be shifted beyond the max model length,
    we therefore have to wait with scheduling.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 49, max tokens = 17, step joining = 0
            * 1: len = 70, max tokens = 17, step joining = 0
    """

    seqs_max_tokens = [17, 17]
    prompts_lengths = [49, 70]
    steps_add_reqs = [0, 0]

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {
            # Prefill sequence 0
            # total blocks in use: 1
            "step": 1,
            "tkv": 49,  # prompt len
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        # Here we cannot schedule sequence 1. By shifting sequence 0 by
        # 1 block its max tkv would exceed the max model length
        {
            # Decode sequence 0
            # total blocks in use: 1 (writing into right pads)
            "step": 2,
            "tkv": 50,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Prefill sequence 1, tkv large enough to prefill w/o tkv shift
            # total blocks in use: 1 + 2
            "step": 17,
            # add 64 to tkv of seq 0 (64) to have it in the same block as seq 1
            "tkv": 128,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            # 2 + 2 (prefill (2 block) + 17 decodes in the last block)
            "n_used_blocks": 3,
        },
        {
            # Decode sequences 0 and 1
            # Sequence 0 finishes
            "step": 18,
            # remove left padding of seq 0, and keep its tkv in the same block
            # as seq 1: 129 - 64 = 65
            # tkv of seq 1 is now max
            "tkv": 71,
            "waiting": [],
            "running": ["1"],
            "request_outputs": ["1", "0"],
            "finished_requests": ["0"],
            "n_used_blocks": 4,  # seq 0 needs another block for the last token
        },
        {
            # Decode sequence 1
            # total blocks in use: 4 - 2 = 2
            "step": 19,
            "tkv": 72,
            "waiting": [],
            "running": ["1"],
            "request_outputs": ["1"],
            "n_used_blocks": 2,
        },
        {
            # Sequence 1 finishes at step 33
            "step": 33,
            "tkv": 86,
            "waiting": [],
            "running": [],
            "request_outputs": ["1"],
            "finished_requests": ["1"],
            "n_used_blocks": 2,
        },
        {
            # Tkv should be cleared one step later
            "step": 34,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    check_scheduler_inference_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        seqs_max_tokens=seqs_max_tokens,
        prompts_lengths=prompts_lengths,
        steps_add_reqs=steps_add_reqs,
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        use_cb=False,
        max_num_batched_tokens=max_num_batched_tokens,
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [128])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_requests_exceed_batch_tkv_limit(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    set_random_seed,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where a request cannot be scheduled right away as the
    max batch x tkv limit, e.g the volumetric limit, is exceeded.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 1: len = 64, max tokens = 2, step joining = 0
            * 2: len = 65, max tokens = 2, step joining = 0
    """

    seqs_max_tokens = [2, 2]
    prompts_lengths = [64, 65]
    steps_add_reqs = [0, 0]
    # total number of blocks needed if scheduled together: (1 + 1)+(1 + 1) = 4
    # note that as not scheduled together, we only needs 2 blocks here
    # needs 2 * (64 + 1) = 2 * 65 = 130
    max_batch_tkv_limit = 129  # not big enough

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {
            # Prefill sequence 0
            # total blocks in use: 1
            "step": 1,
            "tkv": 64,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        # Note: we cannot prefill seq 1 as the volumetric limit
        # max_batch_tkv_limit is exceeded: 129 < 130
        # -> cond5 in can_schedule() is False
        {
            # Decode sequence 0
            # Sequence 0 finishes at step 2
            # total blocks in use: 2
            "step": 2,
            "tkv": 65,
            "waiting": ["1"],
            "running": [],
            "request_outputs": ["0"],
            "finished_requests": ["0"],
            "n_used_blocks": 2,
        },
        {
            # Prefill sequence 1
            # total blocks in use: 2
            "step": 3,
            "tkv": 65,
            "waiting": [],
            "running": ["1"],
            "request_outputs": ["1"],
            "n_used_blocks": 2,  # 2 - 2 + 2
        },
        {
            # Decode sequence 1
            # Sequence 1 finishes at step 4
            # total blocks in use: 3
            "step": 4,
            "tkv": 66,
            "waiting": [],
            "running": [],
            "request_outputs": ["1"],
            "finished_requests": ["1"],
            "n_used_blocks": 2,
        },
        {
            # Tkv should be cleared one step later
            # total blocks in use: 2 - 2 = 0
            "step": 5,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    check_scheduler_inference_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        seqs_max_tokens=seqs_max_tokens,
        prompts_lengths=prompts_lengths,
        steps_add_reqs=steps_add_reqs,
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_batch_tkv_limit=max_batch_tkv_limit,
        use_cb=False,
        max_num_batched_tokens=max_num_batched_tokens,
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [514, 1024])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_single_cp_prefill(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    set_random_seed,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario to test the most basic execution of chunked scheduling:
    a single prompts larger than the chunk size.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 1
            * 0: len = 512, max tokens = 1, step joining = 0
    """
    # max_model_len=514 tests an edge case in the scheduler, but does not work
    # on sendnn
    if backend == "sendnn" and max_model_len == 514:
        pytest.skip("sendnn backend with 514 context length will not work")
    if backend != "sendnn" and max_model_len == 1024:
        pytest.skip("skipping 1024 context length test case for CPU to save test time")

    seqs_max_tokens = [2]
    prompts_lengths = [512]
    steps_add_reqs = [0]

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {
            # Prefill sequence 0 chunk 0
            "step": 1,
            "tkv": 512,
            "waiting": [],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 8,
        },
        {
            # Prefill sequence 0 chunk 1
            "step": 2,
            "tkv": 512,
            "waiting": [],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 8,
        },
        {
            # Prefill sequence 0 chunk 2
            # total blocks in use: 6
            "step": 3,
            "tkv": 512,
            "waiting": [],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 8,
        },
        {
            # Prefill sequence 0 chunk 3
            # total blocks in use: 8
            "step": 4,
            "tkv": 512,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 8,
        },
        {
            # Decode sequence 0
            # seq 0 finishes in this step
            "step": 5,
            "tkv": 513,
            "waiting": [],
            "running": [],
            "request_outputs": ["0"],
            "finished_requests": ["0"],
            "n_used_blocks": 9,
        },
        {
            # Tkv should be cleared one step later
            "step": 6,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    check_scheduler_inference_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        seqs_max_tokens=seqs_max_tokens,
        prompts_lengths=prompts_lengths,
        steps_add_reqs=steps_add_reqs,
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        use_cb=False,
        random_prompts=True,
        max_num_batched_tokens=max_num_batched_tokens,
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [2048])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_cp_prefill_interleave1(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    set_random_seed,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where two sequences are scheduled from the beginning
    and the shorter sequence gets scheduled first. After a couple of
    iterations the interleaving of requests starts.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 1
            * 0: len = 10,  max tokens = 8, step joining = 0
            * 1: len = 512, max tokens = 4, step joining = 0
    """

    seqs_max_tokens = [8, 4]
    prompts_lengths = [10, 512]
    steps_add_reqs = [0, 0]

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {
            # Request 0 is prefilled with a single chunk
            # Token 0 is generated
            "step": 1,
            "tkv": 10,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Request 0 starts decoding.
            # Request 1 can't be scheduled do to a restriction on
            # consecutive prefills. Token 1 is generated
            "step": 2,
            "tkv": 11,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Chunk 0 of request 1 prefill
            "step": 3,
            "tkv": 11,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Decode 2 of request 0.
            "step": 4,
            "tkv": 12,
            "waiting": [],
            "running": ["0", "1"],
            "request_outputs": ["0"],
            "n_used_blocks": 9,
        },
        {
            # Chunk 1 of request 1 prefill
            # tkv of decode batch (tkv not updated until last chunk)
            "step": 5,
            "tkv": 12,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Decode 3 of request 0.
            "step": 6,
            "tkv": 13,
            "waiting": [],
            "running": ["0", "1"],
            "request_outputs": ["0"],
            "n_used_blocks": 9,
        },
        {
            # Chunk 2 of request 1 prefill
            # tkv of decode batch (tkv not updated until last chunk)
            "step": 7,
            "tkv": 13,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Decode 4 of request 0.
            "step": 8,
            "tkv": 14,
            "waiting": [],
            "running": ["0", "1"],
            "request_outputs": ["0"],
            "n_used_blocks": 9,
        },
        {
            # Chunk 3 of request 1 prefill.
            # First token is generated
            # tkv updated for last chunk
            "step": 9,
            "tkv": 512,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 9,
        },
        {
            # Decode 5 of request 0.
            # Decode 1 of request 1.
            "step": 10,
            "tkv": 527,  # tkv jump
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 10,
        },
        {
            # Decode 6 of request 0.
            # Decode 2 of request 1.
            "step": 11,
            "tkv": 528,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 10,
        },
        {
            # Decode 7 of request 0.
            # Decode 3 of request 1.
            # Requests are done
            "step": 12,
            "tkv": 529,
            "waiting": [],
            "running": [],
            "finished_requests": ["1", "0"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 10,
        },
        {
            # Tkv should be cleared one step later
            "step": 13,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    check_scheduler_inference_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        seqs_max_tokens=seqs_max_tokens,
        prompts_lengths=prompts_lengths,
        steps_add_reqs=steps_add_reqs,
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        use_cb=False,
        random_prompts=True,
        max_num_batched_tokens=max_num_batched_tokens,
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [2048])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_cp_prefill_no_interleave(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    set_random_seed,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Same as test_cp_prefill_interleave1 but with interleaving disabled

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 1
            * 0: len = 10,  max tokens = 8, step joining = 0
            * 1: len = 512, max tokens = 4, step joining = 0
    """

    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    seqs_max_tokens = [8, 4]
    prompts_lengths = [10, 512]
    steps_add_reqs = [0, 0]

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {
            # Request 0 is prefilled with a single chunk
            # Token 0 is generated
            "step": 1,
            "tkv": 10,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Chunk 0 of request 1 prefill
            # tkv of decode batch (tkv not updated until last chunk)
            "step": 2,
            "tkv": 10,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Chunk 1 of request 1 prefill
            # tkv of decode batch (tkv not updated until last chunk)
            "step": 3,
            "tkv": 10,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Chunk 2 of request 1 prefill
            # tkv of decode batch (tkv not updated until last chunk)
            "step": 4,
            "tkv": 10,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Chunk 3 of request 1 prefill.
            # First token is generated
            # tkv updated for last chunk
            "step": 5,
            "tkv": 512,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 9,
        },
        {
            # Both requests start decoding.
            "step": 6,
            "tkv": 523,  # tkv jump
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 10,
        },
        {
            # Decode 2
            "step": 7,
            "tkv": 524,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 10,
        },
        {
            # Decode 3
            "step": 8,
            "tkv": 525,
            "waiting": [],
            "running": ["0"],
            "finished_requests": ["1"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 10,
        },
        {
            # Decode 4 of request 0.
            "step": 9,
            "tkv": 14,  # tkv jump
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Decode 5 of request 0.
            "step": 10,
            "tkv": 15,  # tkv jump
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Decode 6 of request 0.
            "step": 11,
            "tkv": 16,  # tkv jump
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Decode 7 of request 0.
            "step": 12,
            "tkv": 17,  # tkv jump
            "waiting": [],
            "running": [],
            "finished_requests": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Tkv should be cleared one step later
            "step": 13,  # with or without interleaving we finish at step 13
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    check_scheduler_inference_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        seqs_max_tokens=seqs_max_tokens,
        prompts_lengths=prompts_lengths,
        steps_add_reqs=steps_add_reqs,
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        use_cb=False,
        random_prompts=True,
        max_num_batched_tokens=max_num_batched_tokens,
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [2048])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_cp_prefill_interleave2(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    set_random_seed,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Same as test_cp_prefill_interleave1 but now the second
    request arrives during decode of step 0

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 1
            * 0: len = 10,  max tokens = 8, step joining = 0
            * 1: len = 512, max tokens = 4, step joining = 3
    """

    seqs_max_tokens = [8, 4]
    prompts_lengths = [10, 512]
    steps_add_reqs = [0, 3]

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {
            # Request 0 is prefilled with a single chunk
            # Token 0 is generated
            "step": 1,
            "tkv": 10,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Decode 1 of request 0.
            "step": 2,
            "tkv": 11,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Decode 2 of request 0.
            "step": 3,
            "tkv": 12,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        {
            # Chunk 0 of request 1 prefill
            # tkv of decode batch (tkv not updated until last chunk)
            "step": 4,
            "tkv": 12,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Decode 3 of request 0.
            "step": 5,
            "tkv": 13,
            "waiting": [],
            "running": ["0", "1"],
            "request_outputs": ["0"],
            "n_used_blocks": 9,
        },
        {
            # Chunk 1 of request 1 prefill
            # tkv of decode batch (tkv not updated until last chunk)
            "step": 6,
            "tkv": 13,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Decode 4 of request 0.
            "step": 7,
            "tkv": 14,
            "waiting": [],
            "running": ["0", "1"],
            "request_outputs": ["0"],
            "n_used_blocks": 9,
        },
        {
            # Chunk 2 of request 1 prefill
            # tkv of decode batch (tkv not updated until last chunk)
            "step": 8,
            "tkv": 14,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
        },
        {
            # Decode 5 of request 0.
            "step": 9,
            "tkv": 15,
            "waiting": [],
            "running": ["0", "1"],
            "request_outputs": ["0"],
            "n_used_blocks": 9,
        },
        {
            # Chunk 3 of request 1 prefill.
            # First token is generated
            # tkv updated for last chunk
            "step": 10,
            "tkv": 512,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 9,
        },
        {
            # Decode 6 of request 0.
            # Decode 1 of request 1.
            "step": 11,
            "tkv": 528,  # tkv jump
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 10,
        },
        {
            # Decode 7 of request 0.
            # Decode 2 of request 1.
            "step": 12,
            "tkv": 529,
            "waiting": [],
            "running": ["1"],
            "finished_requests": ["0"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 10,
        },
        {
            "step": 13,
            "tkv": 515,  # tkv jump
            "waiting": [],
            "running": [],
            "finished_requests": ["1"],
            "request_outputs": ["1"],
            "n_used_blocks": 9,
        },
        {
            # Tkv should be cleared one step later
            "step": 14,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    check_scheduler_inference_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        seqs_max_tokens=seqs_max_tokens,
        prompts_lengths=prompts_lengths,
        steps_add_reqs=steps_add_reqs,
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        use_cb=False,
        random_prompts=True,
        max_num_batched_tokens=max_num_batched_tokens,
    )


# TODO had to move test at the end, having it after test_prefill_tkv_too_big
# was breaking the ordering ("error in test ordering!")
# looks like an issue with sorting the runtime configurations
@pytest.mark.chunked_prefill
@pytest.mark.full_model
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [4])
@pytest.mark.parametrize("max_model_len", [128])  # restricted to violate scheduler condition
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_prefill_tkv_too_big2(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    set_random_seed,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where the requested number of output is too big for current
    tkv value. We need to wait for a previous long prompt request to finish and
    have tkv reduced to a previous block before being able to schedule the
    new request.

    Configuration:
        * max_num_seqs: 4
        * number of prompts: 3
            * 0: len = 20, max tokens = 5, step joining = 0
            * 1: len = 80, max tokens = 3, step joining = 0
            * 2: len = 16, max tokens = 50, step joining = 0
    """

    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    seqs_max_tokens = [5, 3, 50]
    prompts_lengths = [20, 80, 16]
    steps_add_reqs = [0, 0, 0]

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1", "2"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {
            # Prefill sequence 0
            "step": 1,
            "tkv": 20,
            "waiting": ["1", "2"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        # tkv should be updated at the end of the last chunked prefill
        # here we have only one chunk, so it will be updated directly
        {
            # Prefill sequence 1
            "step": 2,
            "tkv": 84,  # 64 (1 block padding) + 20 (prompt of seq 0) = 84
            "waiting": ["2"],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            # 1 + 2 (prefill (2 block) + 3 decodes in the last block)
            "n_used_blocks": 3,
        },
        # Here we cannot schedule sequence 2. Current tkv being in the second
        # block, the number of requested tokens can't fit in the remaining space
        # (64 (full block left padding) + 16 (prompt) + 50 (decode) = 130 > 128)
        {
            # Decode 1 of sequence 0
            # Decode 1 of sequence 1
            "step": 3,
            "tkv": 85,
            "waiting": ["2"],
            "running": ["1", "0"],
            "request_outputs": ["1", "0"],
            "n_used_blocks": 3,
        },
        {
            # Decode 2 of sequence 0
            # Decode 2 of sequence 1
            # Sequence 1 finishes
            "step": 4,
            "tkv": 86,
            "waiting": ["2"],
            "running": ["0"],
            "request_outputs": ["1", "0"],
            "finished_requests": ["1"],
            "n_used_blocks": 3,
        },
        # The tkv value used here is computed before the model forward pass and
        # token sampling of this step. As a result, it does not yet reflect
        # sequences that finish in the current step. In this case, tkv=86 still
        # includes sequence 1, which completes in this step, and this will only
        # be accounted for in the next step. Therefore, sequence 2 cannot be
        # prefilled yet at this point.
        {
            # Decode 3 of sequence 0
            "step": 5,
            "tkv": 23,  # 20 (prompt len) + 3 (decodes) = 23
            "waiting": ["2"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
        },
        # Sequence 2 can be scheduled for prefill, now that tkv is moved back to
        # the first block.
        {
            # Prefill sequence 2
            "step": 6,
            "tkv": 23,
            "waiting": [],
            "running": ["2", "0"],
            "request_outputs": ["2"],
            # 3 - 2 (finished seq 1) + 2 (prefill + 50 decodes in new block)
            "n_used_blocks": 2,
        },
        {
            # Decode 4 of sequence 0
            # Decode 1 of sequence 2
            # Sequence 0 finishes
            "step": 7,
            "tkv": 24,
            "waiting": [],
            "running": ["2"],
            "request_outputs": ["2", "0"],
            "finished_requests": ["0"],
            "n_used_blocks": 2,
        },
        {
            # Decode 2 of sequence 2
            "step": 8,
            "tkv": 18,  # 16 (prompt len) + 2 (decodes) = 18
            "waiting": [],
            "running": ["2"],
            "request_outputs": ["2"],
            "n_used_blocks": 1,
        },
        {
            # Decode 49 of sequence 2
            # Sequence 2 finishes
            "step": 55,
            "tkv": 65,  # 16 (prompt len) + 49 (decodes) = 65
            "waiting": [],
            "running": [],
            "request_outputs": ["2"],
            "finished_requests": ["2"],
            "n_used_blocks": 2,
        },
        {
            # tkv should be cleared one step later
            "step": 56,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    check_scheduler_inference_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        seqs_max_tokens=seqs_max_tokens,
        prompts_lengths=prompts_lengths,
        steps_add_reqs=steps_add_reqs,
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        use_cb=False,
        max_num_batched_tokens=max_num_batched_tokens,
    )



import sys

from transformers import GraniteMoeHybridConfig

# When running this plugin on a Mac, we assume it's for local development
# purposes. However, due to a compatibility issue with vLLM, which overrides
# the Triton module with a placeholder, vLLM may fail to load on macOS. To
# mitigate this issue, we can safely remove the Triton module (if imported)
# and rely on PyTorch to handle the absence of Triton, ensuring fine execution
# in eager mode.
if sys.platform.startswith("darwin"):
    if sys.modules.get("triton"):
        del sys.modules["triton"]

import math
import operator
import os
from typing import TYPE_CHECKING, Union, cast

import torch
from transformers.models.granite import GraniteConfig
from vllm.inputs import ProcessorInputs, PromptType, TokenInputs
from vllm.logger import init_logger

try:
    # pre 0.11.1 compatibility
    from vllm.utils import FlexibleArgumentParser  # ty: ignore[unresolved-import]
except ImportError:
    from vllm.utils.argparse_utils import FlexibleArgumentParser

if TYPE_CHECKING:
    # NB: We can't eagerly import many things from vllm since vllm.config
    # will import this file. These would lead to circular imports
    from vllm.config import ModelConfig, VllmConfig
    from vllm.pooling_params import PoolingParams
    from vllm.sampling_params import SamplingParams
else:
    ModelConfig = None
    VllmConfig = None
    SamplingParams = None
    PoolingParams = None
from vllm.platforms import Platform, PlatformEnum

import vllm_spyre.envs as envs_spyre
from vllm_spyre.compilation_utils import handle_disable_compilation

logger = init_logger(__name__)

THREADING_ENVS = [
    "OMP_NUM_THREADS",
    # "TORCHINDUCTOR_COMPILE_THREADS", # vLLM wants this set to 1
    "DT_PARALLEL_THREADS",  # affects the compilation during warmup
    # set these for good measure
    "OPENBLAS_NUM_THREADS",
    "MKL_NUM_THREADS",
]


# Needed by vllm/model_executor/layers/pooler.py:562
# Copied from vllm/utils/__init__.py
class _StreamPlaceholder:
    def __init__(self):
        self.synchronize = lambda: None


class SpyrePlatform(Platform):
    _enum = PlatformEnum.OOT

    # "spyre" device_name no longer worked due to https://github.com/vllm-project/vllm/pull/16464
    device_name: str = "cpu"
    device_type: str = "cpu"
    # compressed-tensors supported by
    # https://github.com/foundation-model-stack/fms-model-optimizer/blob/main/fms_mo/aiu_addons/__init__.py
    supported_quantization: list[str] = ["gptq", "compressed-tensors"]
    _warmup_shapes: tuple[dict[str, int], ...] | None = None
    _block_size: int = 64  # hardcoded Spyre constraint for now
    # TODO: this `None` is dangerous
    _config: VllmConfig = None  # ty: ignore[invalid-assignment]
    _torch_sendnn_version = None
    # tracks if we are being configured via CLI or LLM() so that we know if
    # default arg parser changes actually have an effect
    _used_with_cli = False

    # Backend for dynamic compilation ops
    # See vllm batched_count_greater_than method
    simple_compile_backend: str = envs_spyre.VLLM_SPYRE_SIMPLE_COMPILE_BACKEND

    # Needed by vllm/model_executor/layers/pooler.py:562
    current_stream = lambda _: _StreamPlaceholder()

    @classmethod
    def get_device_name(cls, device_id: int = 0) -> str:
        return "spyre"

    @classmethod
    def import_kernels(cls) -> None:
        pass  # suppress warning

    @classmethod
    def is_async_output_supported(cls, enforce_eager: bool | None) -> bool:
        """
        Check if the current platform supports async output.
        """
        return False

    @classmethod
    def check_and_update_config(cls, vllm_config: VllmConfig) -> None:
        # 🌶️🌶️🌶️ Patch in our perf logger before the engine is created
        from vllm_spyre.v1.metrics import patch_async_llm_stat_loggers

        patch_async_llm_stat_loggers()

        # In case vllm passes a default vllm_config to us.
        # This happens when get_current_vllm_config is called
        # without setting the vllm config through
        # set_current_vllm_config
        if vllm_config.model_config is None:
            return

        cls._config = vllm_config
        parallel_config = vllm_config.parallel_config
        scheduler_config = vllm_config.scheduler_config
        model_config = vllm_config.model_config
        cache_config = vllm_config.cache_config

        is_decoder = model_config.runner_type == "generate"

        is_pooling = model_config.runner_type == "pooling"

        if not bool(int(os.getenv("VLLM_USE_V1", "1"))):
            raise ValueError("vllm-spyre is only supported with vLLM v1. Please set VLLM_USE_V1=1")
        elif not is_decoder and not is_pooling:
            raise ValueError("Only the 'generate' and 'pooling' runners are supported")

        if parallel_config.worker_cls == "auto":
            parallel_config.worker_cls = "vllm_spyre.v1.worker.spyre_worker.SpyreWorker"

        cls._check_threading_config(parallel_config.world_size)

        # set env vars based on the model
        if is_decoder:
            os.environ["FLEX_OVERWRITE_NMB_FRAME"] = "true"
            os.environ["COMPILATION_MODE"] = "offline_decoder"
        if is_pooling:
            os.environ["FLEX_OVERWRITE_NMB_FRAME"] = "false"
            os.environ["COMPILATION_MODE"] = "offline"

        assert (
            envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL and envs_spyre.VLLM_SPYRE_USE_CB
        ) or not envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL, (
            "Cannot use chunked prefill without continuous batching."
        )

        # enable_prefix_caching will be defaulted to True when used with LLM();
        # only assert if our arg parser default was applied
        if cls._used_with_cli:
            assert (
                cache_config.enable_prefix_caching and envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL
            ) or not cache_config.enable_prefix_caching, (
                "Cannot use prefix caching without chunked prefill."
            )

        if envs_spyre.VLLM_SPYRE_USE_CB and is_decoder:
            if envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL:
                scheduler_config.scheduler_cls = (
                    "vllm_spyre.v1.core.scheduler.ChunkedPrefillSpyreScheduler"
                )
            else:
                scheduler_config.scheduler_cls = (
                    "vllm_spyre.v1.core.scheduler.ContinuousBatchingSpyreScheduler"
                )
            # Overwrite so that vLLM prints our value in the "Initializing a V1
            # LLM engine" log message
            # TODO: With the arg parser defaulting, this can be removed when we
            # only support vllm >= v0.11.1
            if hasattr(scheduler_config, "chunked_prefill_enabled"):
                scheduler_config.chunked_prefill_enabled = envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL  # ty: ignore

            if envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS:
                raise ValueError("Prompt logprobs not supported with continuous batching")
            if (
                vllm_config.model_config.quantization
                and vllm_config.scheduler_config.max_num_seqs == 1
            ):
                raise ValueError("Batch size 1 not supported for fp8 continuous batching.")
        else:
            # Static batching or embedding model.
            # Override --max-num-seqs to the biggest warmup batch size
            # And override --max-model-len to the biggest warmup sequence
            cls._warmup_shapes = None
            spyre_warmup_shapes = cls.get_warmup_shapes(scheduler_config)
            max_batch_size = 0
            max_seq_len = 0
            for shape in spyre_warmup_shapes:
                max_batch_size = max(max_batch_size, shape["batch_size"])
                max_seq_len = max(max_seq_len, shape["prompt_length"] + shape["new_tokens"])

            if envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS and max_batch_size > 1:
                raise ValueError("Prompt logprobs only supported with batch size 1")

            # verify that warmup shapes are not too large
            model_config.get_and_verify_max_len(max_model_len=max_seq_len)

            # override stuff
            model_config.max_model_len = max_seq_len
            scheduler_config.max_num_seqs = max_batch_size

            scheduler_config.scheduler_cls = (
                "vllm_spyre.v1.core.scheduler.StaticBatchingSpyreScheduler"
            )

        # Hardcode some things for granite-3.3-8b-instruct
        if cls.is_granite_3_8b(vllm_config.model_config) or cls.is_granite_4_8b_dense(
            vllm_config.model_config
        ):
            cls.configure_granite_3_8b(vllm_config)

        # v0.14.0+ defaults to async scheduling
        scheduler_config.async_scheduling = False

        # To disable any paged attention ops in the base scheduler, we:
        # - Set the block size (in tokens) to the maximum sequence length
        #       so that the scheduler thinks an entire sequence will fit in
        #       one single block.
        # - Set `max_num_batched_tokens` to the size of a full batch of full
        #       length requests, so that the scheduler will always have token
        #       budget available to schedule a full batch
        if cache_config is not None:
            cache_config.block_size = model_config.max_model_len  # ty: ignore[invalid-assignment]
            if not envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL:
                scheduler_config.max_num_batched_tokens = (
                    model_config.max_model_len * scheduler_config.max_num_seqs
                )
            else:
                # TODO: ideally, this would be user-configurable from CLI/engine
                # args instead of with the internal env var, but that requires a
                # way to detect if value set by vllm or by the user
                if (chunk_len := os.getenv("VLLM_DT_CHUNK_LEN")) is None:
                    os.environ["VLLM_DT_CHUNK_LEN"] = str(scheduler_config.max_num_batched_tokens)
                else:
                    try:
                        chunk_len_int = int(chunk_len)
                    except (ValueError, TypeError) as e:
                        raise Exception("VLLM_DT_CHUNK_LEN must be an integer") from e

                    logger.info(
                        "VLLM_DT_CHUNK_LEN was provided. Overriding max_num_batched_tokens to %d",
                        chunk_len_int,
                    )
                    scheduler_config.max_num_batched_tokens = chunk_len_int

                assert scheduler_config.max_num_batched_tokens % cls._block_size == 0, (
                    "`max_num_batched_tokens` must"
                    f" be divisible by the block size ({cls._block_size}) "
                    "to enable chunked prefill. It was set to "
                    f"`{scheduler_config.max_num_batched_tokens}`. Please "
                    "set `--max-num-batched-tokens` to a number that satisfies "
                    "this constraint."
                )

        logger.info(
            "Configurations for Spyre. max_model_len=%d, max_num_seqs=%d, block_size=%d, "
            "max_num_batched_tokens=%d, enable_chunked_prefill=%r, enable_prefix_caching=%r",
            model_config.max_model_len,
            scheduler_config.max_num_seqs,
            cache_config.block_size,
            scheduler_config.max_num_batched_tokens,
            envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL,
            cache_config.enable_prefix_caching,
        )

        # set env vars for torch_sendnn to consume
        os.environ["VLLM_DT_MAX_CONTEXT_LEN"] = str(vllm_config.model_config.max_model_len)
        if envs_spyre.VLLM_SPYRE_USE_CB and vllm_config.model_config.max_model_len > 32 * 1024:
            logger.warning(
                "Max context length is too big. Currently only 32K (32768) context length is "
                "supported on Spyre for continuous batching. Results might be off!"
            )
        # min value 2 needed for VLLM_DT_MAX_BATCH_SIZE (compiler constraint)
        # Note that we can still have decodes of batch size 1 as the env var
        # only concerns the max batch size.
        os.environ["VLLM_DT_MAX_BATCH_SIZE"] = str(
            max(vllm_config.scheduler_config.max_num_seqs, 2)
        )

        if not os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT"):
            # max product of batch size x tkv supported by the Spyre compiler
            default_max_batch_tkv_limit = (
                vllm_config.model_config.max_model_len * vllm_config.scheduler_config.max_num_seqs
            )

            os.environ["VLLM_DT_MAX_BATCH_TKV_LIMIT"] = str(default_max_batch_tkv_limit)
            logger.info(
                "No model / tensor parallel size specific value for VLLM_DT_MAX_BATCH_TKV_LIMIT "
                "found. Using the default value (max_model_len * max_batch_size): %d",
                default_max_batch_tkv_limit,
            )

        # Compare requested runtime configuration with supported configurations
        # Don't use top-level import to avoid circular import error
        from vllm_spyre.config.runtime_config_validator import validate_runtime_configuration

        warmup_shape_tuples = (
            [(ws["prompt_length"], ws["new_tokens"], ws["batch_size"]) for ws in cls._warmup_shapes]
            if cls._warmup_shapes and not envs_spyre.VLLM_SPYRE_USE_CB
            else None
        )

        validate_runtime_configuration(
            model_config=model_config,
            tp_size=parallel_config.tensor_parallel_size,
            max_model_len=model_config.max_model_len,
            max_num_seqs=scheduler_config.max_num_seqs,
            warmup_shapes=warmup_shape_tuples,
        )

        handle_disable_compilation(vllm_config, is_decoder)

    @classmethod
    def use_all_gather(cls) -> bool:
        """
        Whether to use allgather in LogitsProcessor to gather the logits.
        """
        return True

    @classmethod
    def is_pin_memory_available(cls) -> bool:
        logger.warning("Pin memory is not supported on Spyre.")
        return False

    @classmethod
    def inference_mode(cls):
        """
        Spyre does not support `torch.inference_mode`.
        This allows to fall back to `torch.no_grad` when inference mode is set.
        """
        return torch.no_grad()

    @classmethod
    def get_warmup_shapes(cls, scheduler_config) -> tuple[dict[str, int], ...]:
        if cls._warmup_shapes is not None:
            return cls._warmup_shapes
        # load warmup shapes and sort by "speed"
        wup_prompt_lens = envs_spyre.VLLM_SPYRE_WARMUP_PROMPT_LENS or []
        if not all(pl % 64 == 0 for pl in wup_prompt_lens):
            raise RuntimeError(
                "All values in VLLM_SPYRE_WARMUP_PROMPT_LENS must be multiples of 64."
            )

        wup_batch_sizes = envs_spyre.VLLM_SPYRE_WARMUP_BATCH_SIZES or []
        if len(wup_prompt_lens) != len(wup_batch_sizes):
            raise RuntimeError(
                "The lists in VLLM_SPYRE_WARMUP_PROMPT_LENS and "
                "VLLM_SPYRE_WARMUP_BATCH_SIZES must have equal length"
            )
        if scheduler_config.runner_type == "pooling":
            wup_new_tokens = [0] * len(wup_prompt_lens)
        else:
            wup_new_tokens = envs_spyre.VLLM_SPYRE_WARMUP_NEW_TOKENS or []
            if len(wup_new_tokens) != len(wup_prompt_lens):
                raise RuntimeError(
                    "The lists in VLLM_SPYRE_WARMUP_PROMPT_LENS and "
                    "VLLM_SPYRE_WARMUP_NEW_TOKENS must have equal length"
                )

        logger.info("VLLM_SPYRE_WARMUP_PROMPT_LENS = %s", wup_prompt_lens)
        logger.info("VLLM_SPYRE_WARMUP_NEW_TOKENS = %s", wup_new_tokens)
        logger.info("VLLM_SPYRE_WARMUP_BATCH_SIZES = %s", wup_batch_sizes)

        cls._warmup_shapes = tuple(
            sorted(
                [
                    {"prompt_length": pl, "new_tokens": nt, "batch_size": bs}
                    for pl, nt, bs in zip(wup_prompt_lens, wup_new_tokens, wup_batch_sizes)
                ],
                key=operator.itemgetter("batch_size", "prompt_length"),
            )
        )
        return cls._warmup_shapes

    @classmethod
    def get_block_size(cls) -> int:
        return cls._block_size

    @classmethod
    def supports_v1(cls, model_config: ModelConfig) -> bool:
        """Returns whether the current platform can support v1 for the supplied
        model configuration.
        """
        return True

    @classmethod
    def validate_request(
        cls,
        prompt: PromptType,
        params: Union[SamplingParams, PoolingParams],
        processed_inputs: ProcessorInputs | None = None,
    ) -> None:
        """Raises if this request is unsupported on this platform"""

        # The PoolingParams import is lazy here because it imports vllm.config,
        # which will in turn import this file again.
        from vllm.pooling_params import PoolingParams

        if isinstance(params, PoolingParams):
            # Only validating generation requests for now
            return None

        # Note: Currently prompt logprobs are not supported, therefore
        # envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS is hardcoded to False
        if params.prompt_logprobs is not None and not envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS:
            raise ValueError("Prompt logprobs are currently not supported.")

        # Structured Outputs are not supported yet and cause issues in our
        # scheduler if included in the request
        if params.structured_outputs is not None:
            logger.warning(
                "Structured outputs are currently not supported and "
                "will be stripped from the request."
            )
            params.structured_outputs = None

        if isinstance(prompt, dict) and "prompt_token_ids" in prompt:
            prompt_len = len(prompt["prompt_token_ids"])  # ty: ignore
        elif processed_inputs is not None:
            if "encoder" in processed_inputs:
                raise ValueError("Encoder-decoder models not supported ")
            if "prompt_token_ids" not in processed_inputs:
                # Can't do any extra validation on embedding-only inputs
                return
            prompt_len = len(cast(TokenInputs, processed_inputs)["prompt_token_ids"])
        else:
            # We need a prompt length to do any validation here
            return

        max_tokens = 0
        if params is not None and params.max_tokens is not None:
            max_tokens = params.max_tokens

        if envs_spyre.VLLM_SPYRE_USE_CB:
            # For continuous batching, check if the request is within the max
            # context length. This needs to take the padded prompt length
            # into account.

            # ceil division to pad to next block boundary
            prompt_padding_len = math.ceil(prompt_len / cls._block_size) * cls._block_size
            if prompt_padding_len + max_tokens > cls._config.model_config.max_model_len:
                raise ValueError(
                    "Could not add request: prompt length is "
                    f"{prompt_len} tokens, which gets padded to "
                    f"{prompt_padding_len} tokens, maximum number of output "
                    f"tokens is {max_tokens} tokens, but max model context "
                    f"length is {cls._config.model_config.max_model_len}."
                )
        else:
            # For non-continuous batching, check if the request matches a warmup
            # shape
            assert cls._warmup_shapes is not None, "Warmup shapes must be set"
            if (
                len(
                    cls._get_matching_warmup_shapes(
                        prompt_len=prompt_len,
                        max_tokens=max_tokens,
                        warmup_shapes=cls._warmup_shapes,
                    )
                )
                == 0
            ):
                raise ValueError(
                    "No applicable warmup shape exists for "
                    f"combination of prompt length ({prompt_len} tokens) "
                    "and maximum number of output tokens to be "
                    f"generated ({max_tokens} tokens)"
                )

    @classmethod
    def _get_matching_warmup_shapes(
        cls, prompt_len: int, max_tokens: int, warmup_shapes: tuple[dict[str, int], ...]
    ) -> list[dict[str, int]]:
        """Return the subset of shapes that match this request"""
        return [
            shape
            for shape in warmup_shapes
            if prompt_len <= shape["prompt_length"] and max_tokens <= shape["new_tokens"]
        ]

    # Defined here for testing purposes
    DEFAULT_CHUNK_SIZE = 1024

    @classmethod
    def pre_register_and_update(cls, parser: FlexibleArgumentParser | None = None) -> None:
        if parser is not None:
            # let's us know that defaults were applied to the parser
            cls._used_with_cli = True

            parser.set_defaults(enable_prefix_caching=False)
            # TODO: We don't use the value of the enable_chunked_prefill arg,
            # but setting the default makes logs match our setting.
            # vLLM >= 0.11.1 does not override the arg, so we could remove the
            # env var for v2 if we update our minimum support
            parser.set_defaults(enable_chunked_prefill=envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL)
            if envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL:
                parser.set_defaults(max_num_batched_tokens=cls.DEFAULT_CHUNK_SIZE)

    @classmethod
    def _check_threading_config(cls, worker_count: int):
        """
        Check parallelism configuration to avoid CPU contention

        Libraries that support multi-threading (eg. OpenMP) default to
        parallelism based on the number of CPUs on the host. This can lead to
        CPU contention in containerized deployments especially when process
        forking is involved. This function provides better default behavior.
        """

        # The quay.io/ibm-aiu/spyre-base image includes shell scripts that
        # automatically set OMP_NUM_THREADS to the result of `nproc --all`.
        #
        # vLLM also already has logic around threading to be aware of,
        #  - sets TORCHINDUCTOR_COMPILE_THREADS=1 (https://github.com/vllm-project/vllm/blob/baba0389f7e810a361fff5229ce20c2d5a2b1fac/vllm/env_override.py#L38-L39)
        #  - it will set OMP_NUM_THREADS=1 when using multiple workers (https://github.com/vllm-project/vllm/blob/baba0389f7e810a361fff5229ce20c2d5a2b1fac/vllm/executor/multiproc_worker_utils.py#L304)
        #  - has configurations for OMP thread binding (https://github.com/vllm-project/vllm/blob/baba0389f7e810a361fff5229ce20c2d5a2b1fac/vllm/envs.py#L435-L438)
        #    - the bind attempts to detect NUMA nodes (https://github.com/vllm-project/vllm/blob/baba0389f7e810a361fff5229ce20c2d5a2b1fac/vllm/v1/worker/cpu_worker.py#L111)

        assert worker_count > 0
        # Always print current env for awareness
        env_map = {env: os.getenv(env) for env in THREADING_ENVS}
        logger.info(
            "Initial threading configurations: %s",
            " ".join([f"{env}={value}" for env, value in env_map.items()]),
        )

        # Try to determine the CPU time/cores that we are allocated
        cpu_count: float | None = None
        detection_message = ""

        if (num_cpu := envs_spyre.VLLM_SPYRE_NUM_CPUS) > 0:
            cpu_count = num_cpu
            detection_message = f"VLLM_SPYRE_NUM_CPUS is set to {cpu_count}"
        else:
            try:
                # try to query cgroup CPU limits
                with open("/sys/fs/cgroup/cpu.max") as f:
                    quota_str, period_str = f.read().strip().split()

                if quota_str != "max":
                    quota = int(quota_str)
                    period = int(period_str)
                    cpu_count = quota / period
                    detection_message = f"Detected cgroup CPU limit of {cpu_count}"

            except FileNotFoundError:
                # file may not exist if not running under cgroups v2
                pass
            except Exception as e:
                logger.debug("Error parsing /sys/fs/cgroup/cpu.max to get CPU info", exc_info=e)

            # try psutil to get physical core count
            if cpu_count is None:
                try:
                    import psutil

                    cpu_count = float(psutil.cpu_count(logical=False))
                    detection_message = (
                        f"Detected {cpu_count} physical CPUs from psutil.cpu_count(logical=False)"
                    )
                except ImportError:
                    logger.info("Install psutil to count physical CPU cores")
                    pass
                except Exception as e:
                    logger.debug("Error using psutil", exc_info=e)

            # could try `nproc` here, but it is affected by
            # OMP_NUM_THREADS itself

            # try os.cpu_count() to get node CPU count
            if cpu_count is None and (cpu_count_res := os.cpu_count()) is not None:
                cpu_count = float(cpu_count_res)
                detection_message = f"Detected {cpu_count} CPUs from `os.cpu_count()`"

        # NOTE: math.ceil can output a number for each worker that sums
        # to a total greater than cpu_count.
        cpus_per_worker = math.ceil(cpu_count / worker_count) if cpu_count is not None else None

        thread_warning = (
            "Excessive threads may result in CPU contention. "
            + "Note that each worker processes has its own thread pools."
            if worker_count > 1
            else ""
        )
        failed_detection_message = (
            "Unable to detect available CPUs to validate threading configuration."
        )

        if envs_spyre.VLLM_SPYRE_UPDATE_THREAD_CONFIG:
            if cpus_per_worker is None:
                raise RuntimeError(
                    f"{failed_detection_message} Set VLLM_SPYRE_NUM_CPUS or "
                    "use VLLM_SPYRE_UPDATE_THREAD_CONFIG=0 and configure "
                    "manually."
                )

            for env in THREADING_ENVS:
                os.environ[env] = str(cpus_per_worker)

            logger.info(
                "%s for %d workers. Since VLLM_SPYRE_UPDATE_THREAD_CONFIG is enabled, setting "
                "threading configurations to %d",
                detection_message,
                worker_count,
                cpus_per_worker,
            )
            return

        # In the case that VLLM_SPYRE_UPDATE_THREAD_CONFIG is not enabled,
        # check configs and maybe log a warning
        if cpus_per_worker is None:
            logger.info("%s %s", failed_detection_message, thread_warning)
            return

        def _float_or_0(s: str) -> float:
            try:
                return float(s)
            except ValueError:
                return 0.0

        if any(
            (value is None or _float_or_0(value) > 1.2 * cpus_per_worker)
            for value in env_map.values()
        ):
            logger.warning(
                "%s %s for %d workers. Recommend setting each threading configuration to %d. Set "
                "VLLM_SPYRE_UPDATE_THREAD_CONFIG=1 to do this automatically.",
                thread_warning,
                detection_message,
                worker_count,
                cpus_per_worker,
            )

    def get_max_output_tokens(self, prompt_len: int) -> int:
        """Return the size of biggest ```new_tokens``` of the \
            warmup shapes that fits the prompt length"""
        if self._warmup_shapes is None:
            # ceil division to pad to next block boundary
            padded_prompt_len = math.ceil(prompt_len / self._block_size) * self._block_size
            max_new_tokens = self._config.model_config.max_model_len - padded_prompt_len
            return max_new_tokens

        max_new_tokens = 1
        for shape in self._warmup_shapes:
            if prompt_len <= shape["prompt_length"]:
                max_new_tokens = max(max_new_tokens, shape["new_tokens"])

        return max_new_tokens

    @classmethod
    def _set_env_with_validation(cls, env_var: str, default_value: int) -> None:
        """
        Set an environment variable to a default value if not already set.
        If the user has provided a value, validate it's an integer and raise an error if invalid.
        Warn if it's less than the default.

        Args:
            env_var: The name of the environment variable to set
            default_value: The default value to use (in bytes or appropriate units)

        Raises:
            ValueError: If the user-provided value is not a valid integer
        """
        existing_value = os.getenv(env_var)

        if not existing_value:
            # No existing value, set the default
            os.environ[env_var] = str(default_value)
            logger.info("Using %s = %d", env_var, default_value)
        else:
            # User provided a value, validate it
            try:
                user_value = int(existing_value)
            except ValueError as e:
                raise ValueError(
                    f"{env_var} was set to '{existing_value}' which is not a valid integer"
                ) from e

            # Only warn if user value is less than default
            if user_value < default_value:
                logger.warning(
                    "%s was set to %d which is less than the default of %d",
                    env_var,
                    user_value,
                    default_value,
                )

    @classmethod
    def configure_granite_3_8b(cls, vllm_config: VllmConfig):
        """
        Configure hard coded values for the model
        https://huggingface.co/ibm-granite/granite-3.3-8b-instruct and other dense 8b variants.
        """
        parallel_config = vllm_config.parallel_config

        if parallel_config.world_size != 4:
            # only override configs for TP=4
            return

        # Log once upfront that we detected the model
        logger.info(
            "Granite 8b dense model with tensor parallel size 4 detected. "
            "Applying model-specific configuration overrides."
        )

        tkv_128k = 128 * 1024
        if not os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT"):
            os.environ["VLLM_DT_MAX_BATCH_TKV_LIMIT"] = str(tkv_128k)
            logger.info("Using VLLM_DT_MAX_BATCH_TKV_LIMIT = %d", tkv_128k)
        elif os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT") != str(tkv_128k):
            logger.warning(
                "VLLM_DT_MAX_BATCH_TKV_LIMIT was set to %s, not overriding to default of %d",
                os.getenv("VLLM_DT_MAX_BATCH_TKV_LIMIT"),
                tkv_128k,
            )

        # Set HDMA environment variables with validation
        cls._set_env_with_validation("FLEX_HDMA_P2PSIZE", 256 * 1024 * 1024)  # 256MB
        cls._set_env_with_validation("FLEX_HDMA_COLLSIZE", 32 * 1024 * 1024)  # 32MB

        # Override the total number of KV cache blocks based on what we know
        # will fit. (Unless user already set `--num-gpu-blocks-override`)
        # TODO: remove this once we have correct free memory info available
        if cls.sendnn_configured() and ((0, 0, 0) < cls.sendnn_version() < (1, 0, 3)):
            # Older versions of torch_sendnn use the previous override of ~2k
            # blocks.
            # NB: A version of (0, 0, 0) means that the version of torch_sendnn
            # could not be determined, and we assume this means we have a dev
            # install of newer code.
            blocks_override = 2080
        else:
            # If torch_sendnn is not configured or we have a newer torch_sendnn
            # install, use the newer 8k override.
            blocks_override = 8192

        if vllm_config.cache_config.num_gpu_blocks_override is None:
            vllm_config.cache_config.num_gpu_blocks_override = blocks_override
            logger.info("Overriding available KV Cache blocks to %d", blocks_override)
        elif vllm_config.cache_config.num_gpu_blocks_override != blocks_override:
            logger.warning(
                "--num-gpu-blocks-override was set to %d, not using default of %d",
                vllm_config.cache_config.num_gpu_blocks_override,
                blocks_override,
            )

        # hard-coded value for max_num_batched_tokens with chunked prefill
        if (
            envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL
            and envs_spyre.VLLM_SPYRE_DYNAMO_BACKEND == "sendnn"
            and os.getenv("VLLM_DT_CHUNK_LEN") is None
        ):
            logger.info("Setting --max-num-batched-tokens to 1024 for chunked prefill")
            vllm_config.scheduler_config.max_num_batched_tokens = 1024

    @classmethod
    def is_granite_3_8b(cls, model_config: ModelConfig):
        """Returns true if we have a model that looks like
        ibm-granite/granite-3.3-8b-instruct"""
        if not isinstance(model_config.hf_config, GraniteConfig):
            # Not granite 3 at all
            return False

        return (
            model_config.hf_config.num_hidden_layers == 40
            and model_config.hf_config.max_position_embeddings == 131072
            and model_config.hf_config.hidden_size == 4096
            and model_config.hf_config.vocab_size == 49159
            and model_config.hf_config.num_key_value_heads == 8
            and model_config.hf_config.num_attention_heads == 32
        )

    @classmethod
    def is_granite_4_8b_dense(cls, model_config: ModelConfig):
        """Returns true if we have a dense granite 4 model with the same architecture as granite 3.3
        8b"""
        if not isinstance(model_config.hf_config, GraniteMoeHybridConfig):
            # Not granite 4 at all
            return False

        return (
            model_config.hf_config.num_hidden_layers == 40
            and model_config.hf_config.num_experts_per_tok == 0  # dense model
            and model_config.hf_config.max_position_embeddings == 131072
            and model_config.hf_config.hidden_size == 4096
            and model_config.hf_config.vocab_size == 100352
            and model_config.hf_config.num_key_value_heads == 8
            and model_config.hf_config.num_attention_heads == 32
        )

    @classmethod
    def sendnn_configured(cls) -> bool:
        if envs_spyre.VLLM_SPYRE_DYNAMO_BACKEND == "sendnn":
            try:
                from torch_sendnn._version import __version__ as version_str  # ty: ignore[unresolved-import]

                sem_ver = version_str.split("+")[0]
                cls._torch_sendnn_version = tuple(map(int, sem_ver.split(".")))
                return True
            except ImportError as err:
                raise RuntimeError("sendnn backend requires torch_sendnn") from err
        return False

    @classmethod
    def sendnn_version(cls):
        if cls.sendnn_configured():
            return cls._torch_sendnn_version
        return (0, 0, 0)



import asyncio
from contextlib import ExitStack

import pytest
from spyre_util import DecodeWarmupShapes, ModelInfo, get_chicken_soup_prompts, patch_environment
from vllm import PromptType, SamplingParams
from vllm.engine.arg_utils import AsyncEngineArgs
from vllm.engine.async_llm_engine import AsyncLLMEngine
from vllm.sampling_params import RequestOutputKind
from vllm.v1.engine.async_llm import AsyncLLM


async def generate(
    engine: AsyncLLM | AsyncLLMEngine,
    request_id: str,
    prompt: PromptType,
    output_kind: RequestOutputKind,
    max_tokens: int,
    n: int = 1,
) -> tuple[int, str]:
    # Ensure generate doesn't complete too fast for cancellation test.
    await asyncio.sleep(0.2)

    count = 0
    sampling_params = SamplingParams(
        max_tokens=max_tokens,
        ignore_eos=True,
        output_kind=output_kind,
        seed=42,
        n=n,
    )
    async for out in engine.generate(
        request_id=request_id, prompt=prompt, sampling_params=sampling_params
    ):
        num_tokens = sum(len(output.token_ids) for output in out.outputs)
        if output_kind == RequestOutputKind.DELTA:
            count += num_tokens
        else:
            count = num_tokens

        await asyncio.sleep(0.01)

    return count, request_id


@pytest.mark.parametrize("output_kind", [RequestOutputKind.DELTA, RequestOutputKind.FINAL_ONLY])
@pytest.mark.asyncio
async def test_abort(
    model: ModelInfo,
    backend: str,
    mode: str,
    max_model_len: int,
    max_num_seqs: int,
    max_num_batched_tokens: int,
    warmup_shapes: DecodeWarmupShapes,
    output_kind: RequestOutputKind,
    monkeypatch: pytest.MonkeyPatch,
):
    """Test handling of cancelled requests"""
    with monkeypatch.context() as m, ExitStack() as after:
        patch_kwargs = (
            {
                "use_cb": True,
                "warmup_shapes": None,
                "use_chunked_prefill": mode in ["cp", "pc"],
            }
            if mode in ["cb", "cp", "pc"]
            else {
                "use_cb": False,
                "warmup_shapes": warmup_shapes,
            }
        )
        patch_environment(
            **patch_kwargs,
            backend=backend,
            monkeypatch=m,
        )

        # Async LLM API is a little different between v0 and V1
        engine = AsyncLLM.from_engine_args(
            AsyncEngineArgs(
                model=model.name,
                tokenizer=model.name,
                max_model_len=max_model_len,
                max_num_seqs=max_num_seqs,
                max_num_batched_tokens=max_num_batched_tokens if mode in ["cp", "pc"] else None,
                enable_prefix_caching=mode == "pc",
                revision=model.revision,
                tokenizer_revision=model.revision,
            )
        )
        has_unfinished_requests = engine.output_processor.has_unfinished_requests
        after.callback(engine.shutdown)

        # Test structure here mirrors upstream vLLM test_abort:
        # https://github.com/vllm-project/vllm/blob/e6aab5de2999187c6cf0206f2d63ab6d7a0b6964/tests/v1/engine/test_async_llm.py#L160
        NUM_REQUESTS = 15
        NUM_EXPECTED_TOKENS = 5
        REQUEST_IDS_TO_ABORT = range(1, NUM_REQUESTS, 3)
        PARALLEL_SAMPLE_REQ_IDS = range(1, NUM_REQUESTS, 5)

        request_ids = [f"request-{i}" for i in range(NUM_REQUESTS)]

        # Create concurrent requests
        tasks: list[asyncio.Task] = []
        prompt = get_chicken_soup_prompts(1)[0]
        for idx, request_id in enumerate(request_ids):
            max_tokens = NUM_EXPECTED_TOKENS
            n = 2 if idx in PARALLEL_SAMPLE_REQ_IDS else 1
            tasks.append(
                asyncio.create_task(
                    generate(engine, request_id, prompt, output_kind, max_tokens, n)
                )
            )

        # Simulate cancellation from API server client disconnect
        for idx in REQUEST_IDS_TO_ABORT:
            tasks[idx].cancel()
            await asyncio.sleep(0.1)

        # Confirm that requests actually cancelled and that the other requests
        # are not impacted
        for idx, task in enumerate(tasks):
            if idx in REQUEST_IDS_TO_ABORT:
                with pytest.raises(asyncio.CancelledError):
                    await task
            else:
                num_generated_tokens, request_id = await task
                n = 2 if idx in PARALLEL_SAMPLE_REQ_IDS else 1
                expected_tokens = NUM_EXPECTED_TOKENS * n
                assert num_generated_tokens == expected_tokens, (
                    f"{request_id} generated {num_generated_tokens} but expected {expected_tokens}"
                )

        # Make sure all aborted requests were really aborted
        assert not has_unfinished_requests()

        # Confirm that the server is still up and functioning
        request_id = f"request-{REQUEST_IDS_TO_ABORT[0]}"
        task = asyncio.create_task(
            generate(engine, request_id, prompt, output_kind, NUM_EXPECTED_TOKENS)
        )
        num_generated_tokens, request_id = await task
        assert num_generated_tokens == NUM_EXPECTED_TOKENS
        assert not has_unfinished_requests()



import json
from pathlib import Path

import pytest
from spyre_util import ModelInfo, get_chicken_soup_prompts
from vllm import LLM

from vllm_spyre import envs as envs_spyre


@pytest.mark.cpu
@pytest.mark.cb
def test_file_stats_logger(model: ModelInfo, max_model_len, max_num_seqs, tmp_path):
    prompts = get_chicken_soup_prompts(4)

    envs_spyre.override("VLLM_SPYRE_PERF_METRIC_LOGGING_ENABLED", "1")
    envs_spyre.override("VLLM_SPYRE_PERF_METRIC_LOGGING_DIR", str(tmp_path))
    envs_spyre.override("VLLM_SPYRE_USE_CB", "1")
    envs_spyre.override("VLLM_SPYRE_DYNAMO_BACKEND", "eager")

    model = LLM(
        model=model.name,
        revision=model.revision,
        max_model_len=max_model_len,
        max_num_seqs=max_num_seqs,
        disable_log_stats=False,
    )
    model.generate(prompts=prompts)

    assert Path(tmp_path / "request_metrics.jsonl").exists()

    with Path(tmp_path / "request_metrics.jsonl").open() as f:
        for line in f.readlines():
            data = json.loads(line)
            assert "prefill_interrupt_seconds" in data
            assert "e2e_latency_seconds" in data
            assert "timestamp" in data
